# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import operator
from datetime import timedelta
from math import ceil
from typing import List, Optional, Tuple

import click

from lean.click import LeanCommand, ensure_options
from lean.components.config.optimizer_config_manager import NodeType, available_nodes
from lean.container import container
from lean.models.api import QCOptimizationBacktest, QCProject, QCCompileWithLogs, QCFullOrganization
from lean.models.optimizer import OptimizationConstraint, OptimizationExtremum, OptimizationParameter, \
    OptimizationTarget


def _calculate_backtest_count(parameters: List[OptimizationParameter]) -> int:
    """Calculates the number of backtests needed for the given optimization parameters.

    :param parameters: the parameters to optimize
    :return: the number of backtests a grid search on the parameters would require
    """
    steps_per_parameter = [round((p.max - p.min) / p.step) + 1 for p in parameters]
    return int(functools.reduce(operator.mul, steps_per_parameter, 1))


def _calculate_hours(backtest_time: int, backtest_count: int) -> float:
    """Calculates the total number of hours the optimization will take, given only one node is used.

    :param backtest_time: the number of seconds one backtest is expected to take
    :param backtest_count: the number of backtests that need to be ran
    """
    deploy_time = 30
    backtest_cpu_factor = 1.5
    seconds = (deploy_time + backtest_time * backtest_cpu_factor) * backtest_count
    hours = ceil((seconds * 100) / 3600) / 100
    return max(0.1, hours)


def _format_hours(hours: float) -> str:
    """Format a certain number of hours to a string.

    If the number of hours is less than 1 this returns "x minutes".
    If the number of hours is greater than or equal to 1 this returns "x hours".

    :param hours: the number of hours
    :return: the formatted number of hours
    """
    seconds = timedelta(hours=hours).total_seconds()

    if seconds < 60 * 60:
        amount = round(seconds / 60)
        unit = "minute"
    else:
        amount = round(seconds / (60 * 60))
        unit = "hour"

    unit_suffix = "s" if amount != 1 else ""
    return f"{amount:,} {unit}{unit_suffix}"


def _get_backtest_statistic(backtest: QCOptimizationBacktest, target: str) -> float:
    """Returns a statistic of a backtest.

    :param backtest: the backtest to retrieve the statistic from
    :param target: the target statistic to retrieve, must be one of OptimizerConfigManager.available_targets
    :return: the value of the target statistic on the backtest
    """
    if target == "TotalPerformance.PortfolioStatistics.SharpeRatio":
        return backtest.statistics[15]
    elif target == "TotalPerformance.PortfolioStatistics.CompoundingAnnualReturn":
        return backtest.statistics[6]
    elif target == "TotalPerformance.PortfolioStatistics.ProbabilisticSharpeRatio":
        return backtest.statistics[13]
    elif target == "TotalPerformance.PortfolioStatistics.Drawdown":
        return backtest.statistics[7]
    else:
        raise ValueError(f"Target is not supported: {target}")


def _backtest_meets_constraints(backtest: QCOptimizationBacktest, constraints: List[OptimizationConstraint]) -> bool:
    """Returns whether the backtest meets all constraints.

    :param backtest: the backtest to check
    :param constraints: the constraints the backtest has to meet
    :return: True if the backtest meets all constraints, False if not
    """
    optimizer_config_manager = container.optimizer_config_manager()

    for constraint in constraints:
        expression = str(constraint)

        for target, _ in optimizer_config_manager.available_targets:
            expression = expression.replace(target, str(_get_backtest_statistic(backtest, target)))

        if not eval(expression):
            return False

    return True


def _display_estimate(cloud_project: QCProject,
                      finished_compile: QCCompileWithLogs,
                      organization: QCFullOrganization,
                      name: str,
                      strategy: str,
                      target: OptimizationTarget,
                      parameters: List[OptimizationParameter],
                      constraints: List[OptimizationConstraint],
                      node: NodeType,
                      parallel_nodes: int) -> None:
    """Displays the estimated optimization time and cost."""
    api_client = container.api_client()
    estimate = api_client.optimizations.estimate(cloud_project.projectId,
                                                 finished_compile.compileId,
                                                 name,
                                                 strategy,
                                                 target,
                                                 parameters,
                                                 constraints,
                                                 node.name,
                                                 parallel_nodes)

    backtest_count = _calculate_backtest_count(parameters)

    hours = _calculate_hours(estimate.time, backtest_count)
    batch_time = ceil((hours * 100) / parallel_nodes) / 100
    batch_cost = max(0.01, ceil(node.price * hours * 100) / 100)

    logger = container.logger()
    logger.info(f"Estimated number of backtests: {backtest_count:,}")
    logger.info(f"Estimated batch time: {_format_hours(batch_time)}")
    logger.info(f"Estimated batch cost: ${batch_cost:,.2f}")
    logger.info(
        f"Organization balance: {organization.credit.balance:,.0f} QCC (${organization.credit.balance / 100:,.2f})")


@click.command(cls=LeanCommand)
@click.argument("project", type=str)
@click.option("--target",
              type=str,
              help="The target statistic of the optimization")
@click.option("--target-direction",
              type=click.Choice(["min", "max"], case_sensitive=False),
              help="Whether the target must be minimized or maximized")
@click.option("--parameter",
              type=(str, float, float, float),
              multiple=True,
              help="The 'parameter min max step' pairs configuring the parameters to optimize")
@click.option("--constraint",
              type=str,
              multiple=True,
              help="The 'statistic operator value' pairs configuring the constraints of the optimization")
@click.option("--node",
              type=click.Choice([node.name for node in available_nodes], case_sensitive=False),
              help="The node type to run the optimization on")
@click.option("--parallel-nodes",
              type=int,
              help="The number of nodes that may be run in parallel")
@click.option("--name", type=str, help="The name of the optimization (a random one is generated if not specified)")
@click.option("--push",
              is_flag=True,
              default=False,
              help="Push local modifications to the cloud before starting the optimization")
def optimize(project: str,
             target: Optional[str],
             target_direction: Optional[str],
             parameter: List[Tuple[str, float, float, float]],
             constraint: List[str],
             node: Optional[str],
             parallel_nodes: Optional[int],
             name: Optional[str],
             push: bool) -> None:
    """Optimize a project in the cloud.

    PROJECT must be the name or id of the project to optimize.

    An interactive prompt will be shown to configure the optimizer.
    If --target is given the command runs in non-interactive mode.
    In this mode the CLI does not prompt for input and the following options become required:
    --target, --target-direction, --parameter, --node and --parallel-nodes.

    \b
    In non-interactive mode the --parameter option can be provided multiple times to configure multiple parameters:
    - --parameter <name> <min value> <max value> <step size>
    - --parameter my-first-parameter 1 10 0.5 --parameter my-second-parameter 20 30 5

    \b
    In non-interactive mode the --constraint option can be provided multiple times to configure multiple constraints:
    - --constraint "<statistic> <operator> <value>"
    - --constraint "Sharpe Ratio >= 0.5" --constraint "Drawdown < 0.25"

    If the project that has to be optimized has been pulled to the local drive
    with `lean cloud pull` it is possible to use the --push option to push local
    modifications to the cloud before running the optimization.
    """
    logger = container.logger()
    api_client = container.api_client()

    cloud_project_manager = container.cloud_project_manager()
    cloud_project = cloud_project_manager.get_cloud_project(project, push)

    if name is None:
        name = container.name_generator().generate_name()

    cloud_runner = container.cloud_runner()
    finished_compile = cloud_runner.compile_project(cloud_project)

    optimizer_config_manager = container.optimizer_config_manager()
    organization = api_client.organizations.get(cloud_project.organizationId)

    if target is not None:
        ensure_options(["target", "target_direction", "parameter", "node", "parallel_nodes"])

        optimization_strategy = "QuantConnect.Optimizer.Strategies.GridSearchOptimizationStrategy"
        optimization_target = OptimizationTarget(target=optimizer_config_manager.parse_target(target),
                                                 extremum=target_direction)
        optimization_parameters = optimizer_config_manager.parse_parameters(parameter)
        optimization_constraints = optimizer_config_manager.parse_constraints(constraint)

        node = next(n for n in available_nodes if n.name == node)
        if parallel_nodes < node.min_nodes:
            raise RuntimeError(f"The minimum number of parallel nodes for {node.name} is {node.min_nodes}")
        if parallel_nodes > node.max_nodes:
            raise RuntimeError(f"The maximum number of parallel nodes for {node.name} is {node.max_nodes}")

        _display_estimate(cloud_project,
                          finished_compile,
                          organization,
                          name,
                          optimization_strategy,
                          optimization_target,
                          optimization_parameters,
                          optimization_constraints,
                          node,
                          parallel_nodes)
    else:
        optimization_strategy = optimizer_config_manager.configure_strategy(cloud=True)
        optimization_target = optimizer_config_manager.configure_target()
        optimization_parameters = optimizer_config_manager.configure_parameters(cloud_project.parameters, cloud=True)
        optimization_constraints = optimizer_config_manager.configure_constraints()

        while True:
            node, parallel_nodes = optimizer_config_manager.configure_node()

            _display_estimate(cloud_project,
                              finished_compile,
                              organization,
                              name,
                              optimization_strategy,
                              optimization_target,
                              optimization_parameters,
                              optimization_constraints,
                              node,
                              parallel_nodes)

            if click.confirm("Do you want to start the optimization on the selected node type?", default=True):
                break

    optimization = cloud_runner.run_optimization(cloud_project,
                                                 finished_compile,
                                                 name,
                                                 optimization_strategy,
                                                 optimization_target,
                                                 optimization_parameters,
                                                 optimization_constraints,
                                                 node.name,
                                                 parallel_nodes)

    backtests = optimization.backtests.values()
    backtests = [b for b in backtests if b.exitCode == 0]
    backtests = [b for b in backtests if
                 _backtest_meets_constraints(b, optimization_constraints)]

    if len(backtests) == 0:
        logger.info("No optimal parameter combination found, no successful backtests meet all constraints")
        return

    optimal_backtest = sorted(backtests,
                              key=lambda backtest: _get_backtest_statistic(backtest, optimization_target.target),
                              reverse=optimization_target.extremum == OptimizationExtremum.Maximum)[0]

    parameters = ", ".join(f"{key}: {optimal_backtest.parameterSet[key]}" for key in optimal_backtest.parameterSet)
    logger.info(f"Optimal parameters: {parameters}")

    optimal_backtest = api_client.backtests.get(cloud_project.projectId,
                                                optimal_backtest.id)

    logger.info(f"Optimal backtest id: {optimal_backtest.backtestId}")
    logger.info(f"Optimal backtest name: {optimal_backtest.name}")
    logger.info(f"Optimal backtest results:")
    logger.info(optimal_backtest.get_statistics_table())



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Dict, Type, List
from lean.container import container
from lean.models.brokerages.local.local_brokerage import LocalBrokerage
from lean.models.brokerages.local.data_feed import DataFeed
from lean.models import json_modules

all_local_brokerages: List[LocalBrokerage] = []
all_local_data_feeds: List[DataFeed] = []
local_brokerage_data_feeds: Dict[Type[LocalBrokerage],
                                 List[Type[DataFeed]]] = {}

for json_module in json_modules:
    if "local-brokerage" in json_module["type"]:
        all_local_brokerages.append(LocalBrokerage(json_module))
    if "data-queue-handler" in json_module["type"]:
        all_local_data_feeds.append(DataFeed(json_module))

# Remove IQFeed DataFeed for other than windows machines
if not [container.platform_manager().is_host_windows() or os.environ.get("__README__", "false") == "true"]:
    all_local_data_feeds = [
        data_feed for data_feed in all_local_data_feeds if data_feed._id != "IQFeed"]

for local_brokerage in all_local_brokerages:
    local_brokerage_data_feeds[local_brokerage] = all_local_data_feeds



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, List
import click
from lean.click import PathParameter
from lean.models.configuration import Configuration


def get_click_option_type(configuration: Configuration):
    # get type should be a method of configurations class itself.
    # TODO: handle input can inherit type prompt.
    if configuration._config_type == "internal-input":
        return str
    if configuration._input_method == "confirm":
        return bool
    elif configuration._input_method == "choice":
        return click.Choice(configuration._choices, case_sensitive=False)
    elif configuration._input_method == "prompt":
        return configuration.get_input_type()
    elif configuration._input_method == "prompt-password":
        return str
    elif configuration._input_method == "path-parameter":
        return PathParameter(exists=False, file_okay=True, dir_okay=False)


def get_attribute_type(configuration: Configuration):
    # get type should be a method of configurations class itself.
    # TODO: handle input can inherit type prompt.
    if configuration._config_type == "internal-input":
        return str
    if configuration._input_method == "confirm":
        return bool
    elif configuration._input_method == "choice":
        return str
    elif configuration._input_method == "prompt":
        return configuration.get_input_type()
    elif configuration._input_method == "prompt-password":
        return str
    elif configuration._input_method == "path-parameter":
        return str


def get_the_correct_type_default_value(default_lean_config_key: str, default_input_value: str, expected_type: Any,
                                       choices: List[str] = None):
    from lean.commands.live.deploy import _get_default_value
    lean_value = _get_default_value(default_lean_config_key)
    if lean_value is None and default_input_value is not None:
        lean_value = default_input_value
    # This handles backwards compatibility for the old modules.json values.
    if lean_value is not None and type(lean_value) != expected_type and type(lean_value) == bool:
        if choices and "true" in choices and "false" in choices:
            # Backwards compatibility for zeroha-history-subscription.
            lean_value = "true" if lean_value else "false"
        else:
            # Backwards compatibility for tradier-use-sandbox
            lean_value = "paper" if lean_value else "live"
    return lean_value


def get_options_attributes(configuration: Configuration, default_lean_config_key=None):
    options_attributes = {
        "type": get_click_option_type(configuration),
        "help": configuration._help
    }
    default_input_value = configuration._input_default if configuration._is_required_from_user else None
    options_attributes["default"] = lambda: get_the_correct_type_default_value(
        default_lean_config_key, default_input_value, get_attribute_type(configuration),
        configuration._choices if configuration._input_method == "choice" else None)
    return options_attributes


def get_default_key(configuration: Configuration):
    return "job-organization-id" if configuration.is_type_organization_id else configuration._id


def options_from_json(configurations: List[Configuration]):

    def decorator(f):
        for configuration in reversed(configurations):
            long = configuration._id
            name = str(configuration._id).replace('-', '_')
            param_decls = (
                '--' + long,
                name)
            attrs = get_options_attributes(
                configuration, get_default_key(configuration))
            click.option(*param_decls, **attrs)(f)
        return f
    return decorator



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pydantic import BaseModel, ValidationError


class WrappedBaseModel(BaseModel):
    """A version of Pydantic's BaseModel which makes the input data accessible in case of a validation error."""

    def __init__(self, *args, **kwargs) -> None:
        """Creates a new WrappedBaseModel instance.

        :param args: args to pass on to the BaseModel constructor
        :param kwargs: kwargs to pass on to the BaseModel constructor
        """
        try:
            super().__init__(*args, **kwargs)
        except ValidationError as error:
            error.input_value = kwargs
            raise error



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
from typing import Dict, Optional, Any

from lean.components.config.lean_config_manager import LeanConfigManager
from lean.models.api import QCSecurityType
from lean.models.market_hours_database import MarketHoursDatabaseEntry


class MarketHoursDatabase:
    """The MarketHoursDatabase class handles access to the market-hours-database.json file."""

    def __init__(self, lean_config_manager: LeanConfigManager) -> None:
        """Creates a new MarketHoursDatabase instance.

        :param lean_config_manager: the LeanConfigManager instance that is used when retrieving the data directory
        """
        self._lean_config_manager = lean_config_manager
        self._entries: Optional[str, Dict[str, Any]] = None

    def get_entry(self, security_type: QCSecurityType, market: str, ticker: str) -> MarketHoursDatabaseEntry:
        """Reads the market hours database and returns the entry for the given data.

        An error is raised if the market hours database does not contain an entry matching the given data.

        :param security_type: the security type of the data
        :param market: the market of the data
        :param ticker: the ticker of the data
        :return: the market hours database entry for the data
        """
        entries = self._get_all_entries()

        keys_to_check = [f"{security_type.value}-{market}-{ticker.upper()}", f"{security_type.value}-{market}-[*]"]

        for key in keys_to_check:
            if key in entries:
                return MarketHoursDatabaseEntry(**entries[key])

        raise ValueError(f"Could not find entry in market hours database, checked following keys: {keys_to_check}")

    def _get_all_entries(self) -> Dict[str, MarketHoursDatabaseEntry]:
        """Reads the market hours database and returns all unparsed entries by name.

        :return: a dict containing all unparsed market hours database entries by name
        """
        if self._entries is not None:
            return self._entries

        data_dir = self._lean_config_manager.get_data_directory()
        market_hours_database_path = data_dir / "market-hours" / "market-hours-database.json"

        market_hours_database = market_hours_database_path.read_text(encoding="utf-8")
        market_hours_database = re.sub(r"(/\*.*\*/)", "", market_hours_database, flags=re.DOTALL)
        market_hours_database = json.loads(market_hours_database)

        self._entries = market_hours_database["entries"]
        return self._entries



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import traceback
from pathlib import Path
from typing import List, Optional, Dict

from lean.components.api.api_client import APIClient
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.util.logger import Logger
from lean.components.util.project_manager import ProjectManager
from lean.models.api import QCLanguage, QCProject
from lean.models.utils import LeanLibraryReference


class PushManager:
    """The PushManager class is responsible for synchronizing local projects to the cloud."""

    def __init__(self,
                 logger: Logger,
                 api_client: APIClient,
                 project_manager: ProjectManager,
                 project_config_manager: ProjectConfigManager) -> None:
        """Creates a new PushManager instance.

        :param logger: the logger to use when printing messages
        :param api_client: the APIClient instance to use when communicating with the cloud
        :param project_manager: the ProjectManager to use when looking for certain projects
        :param project_config_manager: the ProjectConfigManager instance to use
        """
        self._logger = logger
        self._api_client = api_client
        self._project_manager = project_manager
        self._project_config_manager = project_config_manager
        self._last_file = None
        self._cloud_projects = []

    def push_projects(self, projects_to_push: List[Path], organization_id: Optional[str] = None) -> None:
        """Pushes the given projects from the local drive to the cloud.

        It will also push every library referenced by each project and add or remove references.

        :param projects_to_push: a list of directories containing the local projects that need to be pushed
        :param organization_id: the id of the organization where the project will be pushed to
        """
        if len(projects_to_push) == 0:
            return

        projects_paths = projects_to_push + [library
                                             for project in projects_to_push
                                             for library in self._project_manager.get_project_libraries(project)]
        projects_paths = sorted(projects_paths)

        pushed_projects = {}
        for index, path in enumerate(projects_paths, start=1):
            relative_path = path.relative_to(Path.cwd())
            try:
                self._logger.info(f"[{index}/{len(projects_paths)}] Pushing '{relative_path}'")
                pushed_projects[path] = self._push_project(path, organization_id)
            except Exception as ex:
                self._logger.debug(traceback.format_exc().strip())
                if self._last_file is not None:
                    self._logger.warn(f"Cannot push '{relative_path}' (failed on {self._last_file}): {ex}")
                else:
                    self._logger.warn(f"Cannot push '{relative_path}': {ex}")

        self._update_cloud_library_references(pushed_projects)

    def _update_cloud_library_references(self, projects: Dict[Path, QCProject]) -> None:
        for path, project in projects.items():
            local_libraries_cloud_ids = self._get_local_libraries_cloud_ids(path)
            self._add_new_libraries(project, local_libraries_cloud_ids)
            self._remove_outdated_libraries(project, local_libraries_cloud_ids)

    def _get_local_libraries_cloud_ids(self, project_dir: Path) -> List[int]:
        project_config = self._project_config_manager.get_project_config(project_dir)

        libraries_in_config = project_config.get("libraries", [])
        library_paths = [LeanLibraryReference(**library).path.expanduser().resolve() for library in libraries_in_config]

        local_libraries_cloud_ids = [int(self._project_config_manager.get_project_config(path).get("cloud-id", None))
                                     for path in library_paths]

        return local_libraries_cloud_ids

    def _get_library_name(self, library_cloud_id: int) -> str:
        return self._get_cloud_project(library_cloud_id).name

    def _add_new_libraries(self, project: QCProject, local_libraries_cloud_ids: List[int]) -> None:
        libraries_to_add = [library_id for library_id in local_libraries_cloud_ids if
                            library_id not in project.libraries]

        if len(libraries_to_add) > 0:
            self._logger.info(f"Adding libraries to project {project.name} in the cloud")

        for i, library_cloud_id in enumerate(libraries_to_add, start=1):
            library_name = self._get_library_name(library_cloud_id)
            self._logger.info(f"[{i}/{len(libraries_to_add)}] "
                              f"Adding library {library_name} to project {project.name} in the cloud")
            self._api_client.projects.add_library(project.projectId, library_cloud_id)

    def _remove_outdated_libraries(self, project: QCProject, local_libraries_cloud_ids: List[int]) -> None:
        libraries_to_remove = [library_id for library_id in project.libraries
                               if library_id not in local_libraries_cloud_ids]

        if len(libraries_to_remove) > 0:
            self._logger.info(f"Removing libraries from project {project.name} in the cloud")

        for i, library_cloud_id in enumerate(libraries_to_remove, start=1):
            library_name = self._get_library_name(library_cloud_id)
            self._logger.info(f"[{i}/{len(libraries_to_remove)}] "
                              f"Removing library {library_name} from project {project.name} in the cloud")
            self._api_client.projects.delete_library(project.projectId, library_cloud_id)

    def _push_project(self, project: Path, organization_id: Optional[str]) -> QCProject:
        """Pushes a single local project to the cloud.

        Raises an error with a descriptive message if the project cannot be pushed.

        :param project: the local project to push
        :param organization_id: the id of the organization to push the project to
        """
        project_name = project.relative_to(Path.cwd()).as_posix()

        project_config = self._project_config_manager.get_project_config(project)
        cloud_id = project_config.get("cloud-id")

        # Find the cloud project to push the files to
        if cloud_id is not None:
            # Project has cloud id which matches cloud project, update cloud project
            cloud_project = self._get_cloud_project(cloud_id)
        else:
            # Project has invalid cloud id or no cloud id at all, create new cloud project
            cloud_project = self._api_client.projects.create(project_name,
                                                             QCLanguage[project_config.get("algorithm-language")],
                                                             organization_id)
            self._cloud_projects.append(cloud_project)
            project_config.set("cloud-id", cloud_project.projectId)
            project_config.set("organization-id", cloud_project.organizationId)

            organization_message_part = f" in organization '{organization_id}'" if organization_id is not None else ""
            self._logger.info(f"Successfully created cloud project '{project_name}'{organization_message_part}")

        # Push local files to cloud
        self._push_files(project, cloud_project)

        # Finalize pushing by updating locally modified metadata
        self._push_metadata(project, cloud_project)

        return cloud_project

    def _push_files(self, project: Path, cloud_project: QCProject) -> None:
        """Pushes the files of a local project to the cloud.

        :param project: the local project to push the files of
        :param cloud_project: the cloud project to push the files to
        """
        cloud_files = self._api_client.files.get_all(cloud_project.projectId)
        local_files = self._project_manager.get_source_files(project)
        local_file_names = [local_file.relative_to(project).as_posix() for local_file in local_files]

        for local_file, file_name in zip(local_files, local_file_names):
            self._last_file = local_file

            if "bin/" in file_name or "obj/" in file_name or ".ipynb_checkpoints/" in file_name:
                continue

            file_content = local_file.read_text(encoding="utf-8")
            cloud_file = next(iter([f for f in cloud_files if f.name == file_name]), None)

            if cloud_file is None:
                new_file = self._api_client.files.create(cloud_project.projectId, file_name, file_content)
                self._project_manager.update_last_modified_time(local_file, new_file.modified)
                self._logger.info(f"Successfully created cloud file '{cloud_project.name}/{file_name}'")
            elif cloud_file.content.strip() != file_content.strip():
                new_file = self._api_client.files.update(cloud_project.projectId, file_name, file_content)
                self._project_manager.update_last_modified_time(local_file, new_file.modified)
                self._logger.info(f"Successfully updated cloud file '{cloud_project.name}/{file_name}'")

        # Delete locally removed files in cloud
        files_to_remove = [cloud_file for cloud_file in cloud_files
                           if (not cloud_file.isLibrary and
                               not any(local_file_name == cloud_file.name for local_file_name in local_file_names))]
        for file in files_to_remove:
            self._last_file = Path(file.name)
            self._api_client.files.delete(cloud_project.projectId, file.name)
            self._logger.info(f"Successfully removed cloud file '{cloud_project.name}/{file.name}'")

        self._last_file = None

    def _push_metadata(self, project: Path, cloud_project: QCProject) -> None:
        """Pushes local project description and parameters to the cloud.

        Does nothing if the cloud is already up-to-date.

        :param project: the local project to push the parameters of
        :param cloud_project: the cloud project to push the parameters to
        """
        project_config = self._project_config_manager.get_project_config(project)

        local_description = project_config.get("description", "")
        cloud_description = cloud_project.description

        local_parameters = project_config.get("parameters", {})
        cloud_parameters = {parameter.key: parameter.value for parameter in cloud_project.parameters}

        # Use latest (-1) by default
        local_lean_version = int(project_config.get("lean-engine", "-1"))
        cloud_lean_version = cloud_project.leanVersionId

        local_lean_venv = project_config.get("python-venv", None)
        cloud_lean_venv = cloud_project.leanEnvironment

        update_args = {}

        if local_description != cloud_description:
            update_args["description"] = local_description

        if local_parameters != cloud_parameters:
            update_args["parameters"] = local_parameters

        if (local_lean_version != cloud_lean_version and
           (local_lean_version != -1 or not cloud_project.leanPinnedToMaster)):
            update_args["lean_engine"] = local_lean_version

        # Initially, python-venv is not defined in the config and the default one will be used.
        # After it is changed, in order to use the default one again, it must not be removed from the config,
        # but it should be set to the default env id explicitly instead.
        if local_lean_venv is not None and local_lean_venv != cloud_lean_venv:
            update_args["python_venv"] = local_lean_venv

        if update_args != {}:
            self._api_client.projects.update(cloud_project.projectId, **update_args)
            self._logger.info(f"Successfully updated {' and '.join(update_args.keys())} for '{cloud_project.name}'")

    def _get_cloud_project(self, project_id: int) -> QCProject:
        project = next(iter(p for p in self._cloud_projects if p.projectId == project_id), None)
        if project is None:
            project = self._api_client.projects.get(project_id)
            self._cloud_projects.append(project)

        return project



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import shutil
import site
import subprocess
import sys
from datetime import datetime, timezone
from pathlib import Path
from typing import List, Optional, Union

import pkg_resources

from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.util.logger import Logger
from lean.components.util.path_manager import PathManager
from lean.components.util.platform_manager import PlatformManager
from lean.components.util.xml_manager import XMLManager
from lean.constants import PROJECT_CONFIG_FILE_NAME
from lean.models.api import QCLanguage, QCProject
from lean.models.utils import LeanLibraryReference


class ProjectManager:
    """The ProjectManager class provides utilities for handling a single project."""

    def __init__(self,
                 logger: Logger,
                 project_config_manager: ProjectConfigManager,
                 lean_config_manager: LeanConfigManager,
                 path_manager: PathManager,
                 xml_manager: XMLManager,
                 platform_manager: PlatformManager) -> None:
        """Creates a new ProjectManager instance.

        :param logger: the logger to use to log messages with
        :param project_config_manager: the ProjectConfigManager to use when creating new projects
        :param lean_config_manager: the LeanConfigManager to get the CLI root directory from
        :param path_manager: the path manager to use to handle library paths
        :param xml_manager: the XMLManager to use when working with XML
        :param platform_manager: the PlatformManager used when checking which operating system is in use
        """
        self._logger = logger
        self._project_config_manager = project_config_manager
        self._lean_config_manager = lean_config_manager
        self._path_manager = path_manager
        self._xml_manager = xml_manager
        self._platform_manager = platform_manager

    def find_algorithm_file(self, input: Path) -> Path:
        """Returns the path to the file containing the algorithm.

        Raises an error if the algorithm file cannot be found.

        :param input: the path to the algorithm or the path to the project
        :return: the path to the file containing the algorithm
        """
        if input.is_file():
            return input

        for file_name in ["main.py", "Main.cs"]:
            target_file = input / file_name
            if target_file.exists():
                return target_file

        raise ValueError("The specified project does not contain a main.py or Main.cs file")

    def get_project_by_id(self, local_id: int) -> Path:
        """Finds a project by its local id.

        Raises an error if a project with the given local id cannot be found.

        :param local_id: the local id of the project
        :return: the path to the directory containing the project with the given local id
        """
        directories = [self._lean_config_manager.get_cli_root_directory()]
        while len(directories) > 0:
            directory = directories.pop(0)

            config_file = directory / PROJECT_CONFIG_FILE_NAME
            if config_file.is_file():
                if self._project_config_manager.get_local_id(directory) == local_id:
                    return directory
            else:
                directories.extend(d for d in directory.iterdir() if d.is_dir())

        raise RuntimeError(f"Project with local id '{local_id}' does not exist")

    def get_source_files(self, directory: Path) -> List[Path]:
        """Returns the paths of all the source files in a directory.

        :param directory: the path to the directory to get the source files of
        :return: the list of source files in the given project directory
        """
        source_files = []

        for obj in directory.iterdir():
            if obj.is_dir():
                if obj.name in ["bin", "obj", ".ipynb_checkpoints", "backtests", "live", "optimizations"]:
                    continue

                source_files.extend(self.get_source_files(obj))

            if obj.suffix not in [".py", ".cs", ".ipynb"]:
                continue

            source_files.append(obj)

        return source_files

    def update_last_modified_time(self, local_file_path: Path, cloud_timestamp: datetime) -> None:
        """Updates the last modified time of a local path to that of the cloud counterpart.

        :param local_file_path: the path to the local file to update the last modified time of
        :param cloud_timestamp: the last modified time of the counterpart of the local file in the cloud
        """
        # Timestamps are stored in UTC in the cloud, but utime() requires them in the local timezone
        time = cloud_timestamp.replace(tzinfo=timezone.utc).astimezone(tz=None)
        time = round(time.timestamp() * 1e9)
        os.utime(local_file_path, ns=(time, time))

    def copy_code(self, project_dir: Path, output_dir: Path) -> None:
        """Copies the source code of a project to another directory.

        :param project_dir: the directory of the project
        :param output_dir: the directory to copy the code to
        """
        output_dir.mkdir(parents=True, exist_ok=True)

        for source_file in self.get_source_files(project_dir):
            target_file = output_dir / source_file.relative_to(project_dir)
            target_file.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(source_file, target_file)

    def create_new_project(self, project_dir: Path, language: QCLanguage) -> None:
        """Creates a new project directory and fills it with some useful files.

        :param project_dir: the directory of the new project
        :param language: the language of the new project
        """
        project_dir.mkdir(parents=True, exist_ok=True)

        project_config = self._project_config_manager.get_project_config(project_dir)
        project_config.set("algorithm-language", language.name)
        project_config.set("parameters", {})
        project_config.set("description", "")

        if language == QCLanguage.Python:
            self._generate_python_library_projects_config()
            self._generate_vscode_python_config(project_dir)
            self._generate_pycharm_config(project_dir)
        else:
            self._generate_vscode_csharp_config(project_dir)
            self._generate_csproj(project_dir)
            self.generate_rider_config()

    def delete_project(self, project_dir: Path) -> None:
        """Deletes a project directory.

        Raises an error if the project directory does not exist.

        :param project_dir: the directory of the project to delete
        """
        try:
            shutil.rmtree(project_dir)
        except FileNotFoundError:
            raise RuntimeError(f"Failed to delete project. Could not find the specified path {project_dir}.")

    def get_projects_by_name_or_id(self, cloud_projects: List[QCProject],
                                   project: Optional[Union[str, int]]) -> List[QCProject]:
        """Returns a list of all the projects in the cloud that match the given name or id.

        :param cloud_projects: all projects fetched from the cloud
        :param project: the name or id of the project
        :return: a list of all the projects in the cloud that match the given name or id
        """
        search_by_id = isinstance(project, int)

        if project is not None:
            project_path = Path(project).as_posix() if not search_by_id else None
            projects = [cloud_project for cloud_project in cloud_projects
                        if (search_by_id and cloud_project.projectId == project or
                            not search_by_id and Path(cloud_project.name).as_posix() == project_path)]

            if len(projects) == 0:
                raise RuntimeError("No project with the given name or id exists in the cloud")
        else:
            projects = cloud_projects

        return projects

    def get_project_libraries(self, project_dir: Path) -> List[Path]:
        """Returns a list of all the libraries referenced by the given project.

        It will also recursively get all the libraries referenced by each library.

        :param project_dir: Path to the project
        :return List of all the libraries referenced by the given project
        """
        return self._get_project_libraries(project_dir)

    def _get_project_libraries(self, project_dir: Path, seen_projects: List[Path] = None) -> List[Path]:
        """Returns a list of all the libraries referenced by the given project.

        This is a helper method to recurse the libraries and get their dependencies as well.

        :param project_dir: Path to the project
        :param seen_projects: List of paths already seen, which serves as recursion stop criteria
        :return List of all the libraries referenced by the given project
        """
        if seen_projects is None:
            seen_projects = [project_dir]

        project_config = self._project_config_manager.get_project_config(project_dir)
        libraries_in_config = project_config.get("libraries", [])
        libraries = [LeanLibraryReference(**library).path.expanduser().resolve() for library in libraries_in_config]

        referenced_libraries = []
        for library_path in libraries:
            # Avoid infinite recursion
            if library_path in seen_projects:
                continue

            seen_projects.append(library_path)
            referenced_libraries.extend(self._get_project_libraries(library_path, seen_projects))

        libraries.extend(referenced_libraries)

        return list(dict.fromkeys(libraries))

    def restore_csharp_project(self, csproj_file: Path, no_local: bool) -> None:
        """Restores a C# project if requested with the no_local flag and if dotnet is on the user's PATH.

        :param csproj_file: Path to the project's csproj file
        :param no_local: Whether restoring the packages locally must be skipped
        """
        if no_local:
            return

        if shutil.which("dotnet") is None:
            self._logger.info(f"Project {csproj_file.parent} will not be restored because dotnet was not found in PATH")
            return

        project_dir = csproj_file.parent
        self._logger.info(
            f"Restoring packages in '{self._path_manager.get_relative_path(project_dir)}' to provide local autocomplete: {project_dir}, {csproj_file}, {Path.cwd()}")

        process = subprocess.run(["dotnet", "restore", str(csproj_file)], cwd=project_dir)

        if process.returncode != 0:
            raise RuntimeError("Something went wrong while restoring packages, see the logs above for more information")

    def try_restore_csharp_project(self, csproj_file: Path, original_csproj_content: str, no_local: bool) -> None:
        """Restores a C# project if requested with the no_local flag and if dotnet is on the user's PATH.

        :param csproj_file: Path to the project's csproj file
        :param original_csproj_content: The original csproj file content
        :param no_local: Whether restoring the packages locally must be skipped
        """
        try:
            self.restore_csharp_project(csproj_file, no_local)
        except RuntimeError as e:
            self._logger.warn(f"Reverting the changes to '{self._path_manager.get_relative_path(csproj_file)}'")
            csproj_file.write_text(original_csproj_content, encoding="utf-8")
            raise e

    def _generate_python_library_projects_config(self) -> None:
        """Generates the required configuration to enable autocomplete on Python library projects."""
        try:
            cli_root_dir = self._lean_config_manager.get_cli_root_directory()
        except:
            return

        library_dir = cli_root_dir / "Library"
        if not library_dir.is_dir():
            return

        if site.ENABLE_USER_SITE:
            site_packages_dir = site.getusersitepackages()
        else:
            site_packages_dir = site.getsitepackages()[0]

        self._generate_file(Path(site_packages_dir) / "lean-cli.pth", str(library_dir))

    def _generate_vscode_python_config(self, project_dir: Path) -> None:
        """Generates Python interpreter configuration and Python debugging configuration for VS Code.

        :param project_dir: the directory of the new project
        """
        self._generate_file(project_dir / ".vscode" / "settings.json", json.dumps({
            "python.pythonPath": sys.executable,
            "python.languageServer": "Pylance"
        }, indent=4))

        launch_config = {
            "version": "0.2.0",
            "configurations": [
                {
                    "name": "Debug with Lean CLI",
                    "type": "python",
                    "request": "attach",
                    "connect": {
                        "host": "localhost",
                        "port": 5678
                    },
                    "pathMappings": [
                        {
                            "localRoot": "${workspaceFolder}",
                            "remoteRoot": "/LeanCLI"
                        }
                    ]
                }
            ]
        }

        try:
            library_dir = self._lean_config_manager.get_cli_root_directory() / "Library"
            launch_config["configurations"][0]["pathMappings"].append({
                "localRoot": str(library_dir),
                "remoteRoot": "/Library"
            })
        except:
            pass

        self._generate_file(project_dir / ".vscode" / "launch.json", json.dumps(launch_config, indent=4))

    def _generate_pycharm_config(self, project_dir: Path) -> None:
        """Generates Python interpreter configuration and Python debugging configuration for PyCharm.

        :param project_dir: the directory of the new project
        """
        # Generate Python JDK entry for PyCharm Professional and PyCharm Community
        for editor in ["PyCharm", "PyCharmCE"]:
            for directory in self._get_jetbrains_config_dirs(editor):
                self._generate_pycharm_jdk_entry(directory)

        self._generate_file(project_dir / ".idea" / f"{project_dir.name}.iml", """
<?xml version="1.0" encoding="UTF-8"?>
<module type="PYTHON_MODULE" version="4">
  <component name="NewModuleRootManager">
    <content url="file://$MODULE_DIR$" />
    <orderEntry type="jdk" jdkName="Lean CLI" jdkType="Python SDK" />
    <orderEntry type="sourceFolder" forTests="false" />
  </component>
</module>
        """)

        self._generate_file(project_dir / ".idea" / "misc.xml", """
<?xml version="1.0" encoding="UTF-8"?>
<project version="4">
  <component name="ProjectRootManager" version="2" project-jdk-name="Lean CLI" project-jdk-type="Python SDK" />
</project>
        """)

        self._generate_file(project_dir / ".idea" / "modules.xml", f"""
<?xml version="1.0" encoding="UTF-8"?>
<project version="4">
  <component name="ProjectModuleManager">
    <modules>
      <module fileurl="file://$PROJECT_DIR$/.idea/{project_dir.name}.iml" filepath="$PROJECT_DIR$/.idea/{project_dir.name}.iml" />
    </modules>
  </component>
</project>
        """)

        try:
            library_dir = self._lean_config_manager.get_cli_root_directory() / "Library"
            library_dir = f"$PROJECT_DIR$/{os.path.relpath(library_dir, project_dir)}".replace("\\", "/")
            library_mapping = f'<mapping local-root="{library_dir}" remote-root="/Library" />'
        except:
            library_mapping = ""

        self._generate_file(project_dir / ".idea" / "workspace.xml", f"""
<?xml version="1.0" encoding="UTF-8"?>
<project version="4">
  <component name="RunManager" selected="Python Debug Server.Debug with Lean CLI">
    <configuration name="Debug with Lean CLI" type="PyRemoteDebugConfigurationType" factoryName="Python Remote Debug">
      <module name="LEAN" />
      <option name="PORT" value="6000" />
      <option name="HOST" value="localhost" />
      <PathMappingSettings>
        <option name="pathMappings">
          <list>
            <mapping local-root="$PROJECT_DIR$" remote-root="/LeanCLI" />
            {library_mapping}
          </list>
        </option>
      </PathMappingSettings>
      <option name="REDIRECT_OUTPUT" value="true" />
      <option name="SUSPEND_AFTER_CONNECT" value="true" />
      <method v="2" />
    </configuration>
    <list>
      <item itemvalue="Python Debug Server.Debug with Lean CLI" />
    </list>
  </component>
</project>
        """)

    def _generate_pycharm_jdk_entry(self, pycharm_config_dir: Path) -> None:
        """Generates a "Lean CLI" Python JDK entry to PyCharm's internal JDK table.

        When we generate PyCharm's .idea directory we want to tell PyCharm where the Python interpreter is located.
        PyCharm stores this bit of configuration globally, so we find the global location and update it to our needs.

        If PyCharm is not installed yet, we create the configuration anyways.
        Once the user installs PyCharm, it will then automatically pick up the configuration we created in the past.

        :param pycharm_config_dir: the path to the global configuration directory of a PyCharm edition
        """
        # Parse the file containing PyCharm's internal table of Python interpreters
        jdk_table_file = pycharm_config_dir / "options" / "jdk.table.xml"
        if jdk_table_file.exists():
            root = self._xml_manager.parse(jdk_table_file.read_text(encoding="utf-8"))
        else:
            root = self._xml_manager.parse("""
<application>
  <component name="ProjectJdkTable">
  </component>
</application>
            """)

        # Don't do anything if the Lean CLI interpreter entry already exists
        if root.find(".//jdk/name[@value='Lean CLI']") is not None:
            return

        # Add the new JDK entry to the XML tree
        classpath_entries = [Path(p).as_posix() for p in sys.path if p != "" and not p.endswith(".zip")]
        classpath_entries = [f'<root url="{p}" type="simple" />' for p in classpath_entries]
        classpath_entries = "\n".join(classpath_entries)

        component_element = root.find(".//component[@name='ProjectJdkTable']")
        component_element.append(self._xml_manager.parse(f"""
<jdk version="2">
  <name value="Lean CLI" />
  <type value="Python SDK" />
  <version value="Python {sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}" />
  <homePath value="{Path(sys.executable).as_posix()}" />
  <roots>
    <classPath>
      <root type="composite">
        {classpath_entries}
        <root url="file://$APPLICATION_HOME_DIR$/plugins/python/helpers/python-skeletons" type="simple" />
        <root url="file://$APPLICATION_HOME_DIR$/plugins/python/helpers/typeshed/stdlib/3" type="simple" />
        <root url="file://$APPLICATION_HOME_DIR$/plugins/python/helpers/typeshed/stdlib/2and3" type="simple" />
        <root url="file://$APPLICATION_HOME_DIR$/plugins/python/helpers/typeshed/third_party/3" type="simple" />
        <root url="file://$APPLICATION_HOME_DIR$/plugins/python/helpers/typeshed/third_party/2and3" type="simple" />
      </root>
    </classPath>
    <sourcePath>
      <root type="composite" />
    </sourcePath>
  </roots>
</jdk>
        """))

        # Save the modified XML tree
        self._generate_file(jdk_table_file, self._xml_manager.to_string(root))

    def _generate_vscode_csharp_config(self, project_dir: Path) -> None:
        """Generates C# debugging configuration for VS Code.

        :param project_dir: the directory of the new project
        """
        self._generate_file(project_dir / ".vscode" / "launch.json", """
{
    "version": "0.2.0",
    "configurations": [
        {
            "name": "Debug with Lean CLI",
            "request": "attach",
            "type": "coreclr",
            "processId": "1",
            "pipeTransport": {
                "pipeCwd": "${workspaceRoot}",
                "pipeProgram": "docker",
                "pipeArgs": ["exec", "-i", "lean_cli_vsdbg"],
                "debuggerPath": "/root/vsdbg/vsdbg",
                "quoteArgs": false
            },
            "logging": {
                "moduleLoad": false
            }
        }
    ]
}
        """)

    def _generate_csproj(self, project_dir: Path) -> None:
        """Generates a .csproj file for the given project and returns the path to it.

        :param project_dir: the path of the new project
        """
        self._generate_file(project_dir / f"{project_dir.name}.csproj", self.get_csproj_file_default_content())

    def generate_rider_config(self) -> None:
        """Generates C# debugging configuration for Rider."""
        ssh_dir = Path("~/.lean/ssh").expanduser()

        # Add SSH keys to .lean/ssh if necessary
        if not ssh_dir.exists():
            ssh_dir.mkdir(parents=True)
            for name in ["key", "key.pub", "README.md"]:
                with (ssh_dir / name).open("wb+") as file:
                    file.write(pkg_resources.resource_string("lean", f"ssh/{name}"))

        # Find Rider's global configuration directory
        for directory in self._get_jetbrains_config_dirs("Rider"):
            self._generate_rider_debugger_entry(directory, ssh_dir)

    def _generate_rider_debugger_entry(self, rider_config_dir: Path, ssh_dir: Path) -> None:
        """Generates a "root@localhost:2222" remote debugger entry to Rider's internal debugger configuration.

        If Rider is not installed yet, we create the configuration anyways.
        Once the user installs Rider, it will then automatically pick up the configuration we created in the past.

        :param rider_config_dir: the path to the global configuration directory of a Rider edition
        :param ssh_dir: the path to the directory containing the SSH keys
        """
        # Parse the file containing Rider's internal list of remote hosts
        debugger_file = rider_config_dir / "options" / "debugger.xml"
        if debugger_file.exists():
            root = self._xml_manager.parse(debugger_file.read_text(encoding="utf-8"))
        else:
            root = self._xml_manager.parse("""
<application>
    <component name="XDebuggerSettings">
    </component>
</application>
            """)

        component_element = root.find(".//component[@name='XDebuggerSettings']")

        if root.find(".//debuggers") is None:
            component_element.append(self._xml_manager.parse("<debuggers></debuggers>"))
        debuggers = root.find(".//debuggers")

        if debuggers.find(".//debugger[@id='dotnet_debugger']") is None:
            debuggers.append(self._xml_manager.parse('<debugger id="dotnet_debugger"></debugger>'))
        dotnet_debugger = debuggers.find(".//debugger[@id='dotnet_debugger']")

        if dotnet_debugger.find(".//configuration") is None:
            dotnet_debugger.append(self._xml_manager.parse("<configuration></configuration>"))
        configuration = dotnet_debugger.find(".//configuration")

        if configuration.find(".//option[@name='sshCredentials']") is None:
            configuration.append(self._xml_manager.parse('<option name="sshCredentials"></option>'))
        ssh_credentials = configuration.find(".//option[@name='sshCredentials']")

        required_value = f"&lt;credentials HOST=&quot;localhost&quot; PORT=&quot;2222&quot; USERNAME=&quot;root&quot; PRIVATE_KEY_FILE=&quot;{ssh_dir.as_posix()}/key&quot; USE_KEY_PAIR=&quot;true&quot; USE_AUTH_AGENT=&quot;false&quot; /&gt;"

        # Don't do anything if the required entry already exists
        for option in ssh_credentials.findall(f".//option"):
            if option.get("value") == required_value.replace("&lt;", "<").replace("&gt;", ">").replace("&quot;", '"'):
                return

        # Add the new entry to the XML tree
        ssh_credentials.append(self._xml_manager.parse(f'<option value="{required_value}"/>'))

        # Save the modified XML tree
        self._generate_file(debugger_file, self._xml_manager.to_string(root))

    def _generate_file(self, file: Path, content: str) -> None:
        """Writes to a file, which is created if it doesn't exist yet, and normalized the content before doing so.

        :param file: the file to write to
        :param content: the content to write to the file
        """
        file.parent.mkdir(parents=True, exist_ok=True)
        with file.open("w+", encoding="utf-8") as file:
            file.write(content.strip() + "\n")

    def _get_jetbrains_config_dirs(self, editor_name: str) -> List[Path]:
        """Returns the paths to the global configuration directories for all installed editions of a JetBrains IDE.

        :param editor_name: the name of the JetBrains IDE
        :return: the paths to the directories holding the global configuration for the specified IDE
        """
        # Find JetBrains' root directory containing the global configuration directories for all installed IDEs
        # See https://www.jetbrains.com/help/pycharm/project-and-ide-settings.html#ide_settings
        if self._platform_manager.is_host_windows():
            root_dir = Path("~/AppData/Roaming/JetBrains").expanduser()
        elif self._platform_manager.is_host_macos():
            root_dir = Path("~/Library/Application Support/JetBrains").expanduser()
        else:
            root_dir = Path("~/.config/JetBrains").expanduser()

        if not root_dir.exists():
            root_dir.mkdir(parents=True)

        # Find the global config directories for the given IDE
        directories = []

        for path in root_dir.iterdir():
            if not path.is_dir() or not path.name.startswith(editor_name):
                continue

            suffix = path.name.replace(editor_name, "")
            if len(suffix) > 0 and not suffix[0].isdigit():
                continue

            directories.append(path)

        if len(directories) == 0:
            directories.append(root_dir / editor_name)

        return directories

    def get_cloud_project_libraries(self,
                                    cloud_projects: List[QCProject],
                                    project: QCProject,
                                    seen_libraries: List[int] = None) -> List[QCProject]:
        """Gets the libraries referenced by the project and its dependencies from the given cloud projects.

        It recursively gets every Lean CLI library referenced by the project
        and the ones referenced by those libraries as well.

        :param cloud_projects: the cloud projects list to search in.
        :param project: the starting point for the libraries gathering.
        :param seen_libraries: list of seen library IDs to avoid infinite recursion.
        """
        if seen_libraries is None:
            seen_libraries = []

        libraries = [cloud_project
                     for library_id in project.libraries
                     for cloud_project in cloud_projects if cloud_project.projectId == library_id]

        referenced_libraries = []
        for library in libraries:
            # Avoid infinite recursion
            if library.projectId in seen_libraries:
                continue

            seen_libraries.append(library.projectId)
            referenced_libraries.extend(self.get_cloud_project_libraries(cloud_projects, library, seen_libraries))

        libraries.extend(referenced_libraries)

        return list(set(libraries))

    def get_cloud_projects_libraries(self,
                                     cloud_projects: List[QCProject],
                                     projects: List[QCProject],
                                     seen_projects: List[int] = None) -> List[QCProject]:
        """Gets the libraries referenced by the passed projects and its dependencies from the given cloud projects.

        It recursively gets every Lean CLI library referenced by the passed projects
        and the ones referenced by those libraries as well.

        :param cloud_projects: the cloud projects list to search in.
        :param projects: the starting point list of projects for the libraries gathering.
        :param seen_projects: list of seen project IDs to avoid infinite recursion.
        """
        if seen_projects is None:
            seen_projects = [project.projectId for project in projects]

        libraries = []
        for project in projects:
            libraries.extend(self.get_cloud_project_libraries(cloud_projects, project, seen_projects))

        return list(set(libraries))

    @staticmethod
    def get_csproj_file_default_content() -> str:
        return """
<Project Sdk="Microsoft.NET.Sdk">
    <PropertyGroup>
        <Configuration Condition=" '$(Configuration)' == '' ">Debug</Configuration>
        <Platform Condition=" '$(Platform)' == '' ">AnyCPU</Platform>
        <TargetFramework>net6.0</TargetFramework>
        <OutputPath>bin/$(Configuration)</OutputPath>
        <AppendTargetFrameworkToOutputPath>false</AppendTargetFrameworkToOutputPath>
        <DefaultItemExcludes>$(DefaultItemExcludes);backtests/*/code/**;live/*/code/**;optimizations/*/code/**</DefaultItemExcludes>
        <NoWarn>CS0618</NoWarn>
    </PropertyGroup>
    <ItemGroup>
        <PackageReference Include="QuantConnect.Lean" Version="2.5.*"/>
        <PackageReference Include="QuantConnect.DataSource.Libraries" Version="2.5.*"/>
    </ItemGroup>
</Project>
        """

    @staticmethod
    def get_csproj_file_path(project_dir: Path) -> Path:
        """Gets the path to the csproj file in the project directory.

        :param project_dir: Path to the project directory
        :return: Path to the csproj file in the project directory
        """
        return next((p for p in project_dir.iterdir() if p.name.endswith(".csproj")), None)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from lean.components.api.api_client import APIClient
from lean.components.cloud.pull_manager import PullManager
from lean.components.cloud.push_manager import PushManager
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.util.path_manager import PathManager
from lean.models.api import QCProject


class CloudProjectManager:
    """The CloudProjectManager class is responsible for finding the correct cloud project in cloud commands."""

    def __init__(self,
                 api_client: APIClient,
                 project_config_manager: ProjectConfigManager,
                 pull_manager: PullManager,
                 push_manager: PushManager,
                 path_manager: PathManager) -> None:
        """Creates a new PullManager instance.

        :param api_client: the APIClient instance to use when communicating with the cloud
        :param project_config_manager: the ProjectConfigManager instance to use
        :param pull_manager: the PullManager instance to use
        :param push_manager: the PushManager instance to use
        :param path_manager: the PathManager instance to use when validating paths
        """
        self._api_client = api_client
        self._project_config_manager = project_config_manager
        self._pull_manager = pull_manager
        self._push_manager = push_manager
        self._path_manager = path_manager

    def get_cloud_project(self, input: str, push: bool) -> QCProject:
        """Retrieves the cloud project to use given a certain input and whether the local project needs to be pushed.

        Many cloud commands look like "lean cloud <command> <name or id> --push".
        This method handles parsing the "<name or id> --push" part.
        This method returns the cloud project we think the user wants to use with the given command.

        :param input: the input the user gave, either a local project name, a cloud project name or a cloud id
        :param push: True if the local counterpart of the cloud project needs to be pushed
        :return: the cloud project to use
        """
        # If the given input is a valid project directory, we try to use that project
        local_path = Path.cwd() / input
        if self._project_config_manager.try_get_project_config(local_path, self._path_manager):
            if push:
                self._push_manager.push_projects([local_path])

                cloud_id = self._project_config_manager.get_project_config(local_path).get("cloud-id")
                if cloud_id is None:
                    raise RuntimeError("Something went wrong while pushing the project to the cloud")

                return self._api_client.projects.get(cloud_id)
            else:
                cloud_id = self._project_config_manager.get_project_config(local_path).get("cloud-id")
                if cloud_id is not None:
                    return self._api_client.projects.get(cloud_id)

        # If the given input is not a valid project directory, we look for a cloud project with a matching name or id
        # If there are multiple, we use the first one
        for cloud_project in self._api_client.projects.get_all():
            if str(cloud_project.projectId) != input and cloud_project.name != input:
                continue

            # If the user wants to push and the input doesn't match a local project name, we attempt to push again
            # This may happen if the input is an id instead of a name
            # If the local directory exists, we push it and return the updated cloud project
            if push:
                local_path = self._pull_manager.get_local_project_path(cloud_project)
                if local_path.exists():
                    self._push_manager.push_projects([local_path])
                    return self._api_client.projects.get(cloud_project.projectId)

            return cloud_project

        raise RuntimeError("No project with the given name or id could be found")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from lean.click import LeanCommand
from lean.container import container
from lean.models.errors import AuthenticationError


@click.command(cls=LeanCommand)
def whoami() -> None:
    """Display who is logged in."""
    logger = container.logger()
    api_client = container.api_client()
    cli_config_manager = container.cli_config_manager()

    if cli_config_manager.user_id.get_value() is not None and cli_config_manager.api_token.get_value() is not None:
        try:
            organizations = api_client.organizations.get_all()
            logged_in = True
        except AuthenticationError:
            logged_in = False
    else:
        logged_in = False

    if not logged_in:
        logger.info("You are not logged in")
        return

    personal_organization_id = next(o.id for o in organizations if o.ownerName == "You")
    personal_organization = api_client.organizations.get(personal_organization_id)
    member = next(m for m in personal_organization.members if m.isAdmin)

    logger.info(f"You are logged in as {member.name} ({member.email})")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Optional

import click

from lean.click import LeanCommand, PathParameter
from lean.constants import PROJECT_CONFIG_FILE_NAME
from lean.container import container


@click.command(cls=LeanCommand)
@click.option("--project",
              type=PathParameter(exists=True, file_okay=False, dir_okay=True),
              help="Path to the local project to push (all local projects if not specified)")
@click.option("--organization-id",
              type=str,
              help="ID of the organization where the project will be created in. This is ignored if the project has "
                   "already been created in the cloud")
def push(project: Optional[Path], organization_id: Optional[str]) -> None:
    """Push local projects to QuantConnect.

    This command overrides the content of cloud files with the content of their respective local counterparts.

    This command will delete cloud files which don't have a local counterpart.
    """
    # Parse which projects need to be pushed
    if project is not None:
        project_config_manager = container.project_config_manager()
        project_config = project_config_manager.get_project_config(project)
        if not project_config.file.exists():
            raise RuntimeError(f"'{project}' is not a Lean project")

        projects_to_push = [project]
    else:
        projects_to_push = [p.parent for p in Path.cwd().rglob(PROJECT_CONFIG_FILE_NAME)]

    push_manager = container.push_manager()
    push_manager.push_projects(projects_to_push, organization_id)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from pathlib import Path
from typing import List, Optional

from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.config.storage import Storage


class OutputConfigManager:
    """The OutputConfigManager class manages the configuration of a backtest, optimization or live trading session."""

    def __init__(self, lean_config_manager: LeanConfigManager) -> None:
        """Creates a new OutputConfigManager instance.

        :param lean_config_manager: the LeanConfigManager to get the CLI root directory from
        """
        self._lean_config_manager = lean_config_manager

    def get_output_config(self, output_directory: Path) -> Storage:
        """Returns a Storage instance to get/set the configuration of the contents of an output directory.

        :param output_directory: the path to the project to retrieve the configuration of
        :return: the Storage instance containing the configuration of the given backtest/optimization/live trading
        """
        return Storage(str(output_directory / "config"))

    def get_backtest_id(self, backtest_directory: Path) -> int:
        """Returns the id of a backtest.

        :param backtest_directory: the path to the backtest to retrieve the id of
        :return: the id of the given backtest
        """
        return self._get_id(backtest_directory, 1)

    def get_backtest_name(self, backtest_directory: Path) -> str:
        """Returns the name of a backtest.

        :param backtest_directory: the path to the backtest to retrieve the id of
        :return: the name of the given backtest
        """
        config = self.get_output_config(backtest_directory)

        if config.has("backtest-name"):
            return config.get("backtest-name")

        raise ValueError("Backtest name is not set")

    def get_container_name(self, backtest_directory: Path) -> str:
        """Returns the name of a the docker container lean is running in.

        :param backtest_directory: the path to the backtest to retrieve the id of
        :return: the name of the docker container lean is running in.
        """
        config = self.get_output_config(backtest_directory)

        if config.has("container"):
            return config.get("container")

        raise ValueError("Container name is not set")

    def get_backtest_by_id(self, backtest_id: int, root_directory: Optional[Path] = None) -> Path:
        """Finds the directory of a backtest by its id.

        :param backtest_id: the id of the backtest to get the directory of
        :param root_directory: the directory to search from, defaults to the `lean init` directory
        :return: the output directory of the backtest with the given id
        """
        return self._get_by_id("Backtest", backtest_id, ["backtests/*", "optimizations/*/*"], root_directory)

    def get_optimization_id(self, optimization_directory: Path) -> int:
        """Returns the id of an optimization.

        :param optimization_directory: the path to the optimization to retrieve the id of
        :return: the id of the given optimization
        """
        return self._get_id(optimization_directory, 2)

    def get_optimization_by_id(self, optimization_id: int, root_directory: Optional[Path] = None) -> Path:
        """Finds the directory of an optimization by its id.

        :param optimization_id: the id of the optimization to get the directory of
        :param root_directory: the directory to search from, defaults to the `lean init` directory
        :return: the output directory of the optimization with the given id
        """
        return self._get_by_id("Optimization", optimization_id, ["optimizations/*"], root_directory)

    def get_live_deployment_id(self, live_deployment_directory: Path) -> int:
        """Returns the id of a live deployment.

        :param live_deployment_directory: the path to the live deployment to retrieve the id of
        :return: the id of the given optimization
        """
        return self._get_id(live_deployment_directory, 3)

    def get_live_deployment_by_id(self, live_deployment_id: int, root_directory: Optional[Path] = None) -> Path:
        """Finds the directory of a live deployment by its id.

        :param live_deployment_id: the id of the live deployment to get the directory of
        :param root_directory: the directory to search from, defaults to the `lean init` directory
        :return: the output directory of the live deployment with the given id
        """
        return self._get_by_id("Live deployment", live_deployment_id, ["live/*"], root_directory)

    def _get_id(self, output_directory: Path, prefix: int) -> int:
        config = self.get_output_config(output_directory)

        if config.has("id"):
            return config.get("id")

        new_id = int(str(prefix) + str(random.randint(100_000_000, 999_999_999)))
        config.set("id", new_id)

        return new_id

    def _get_by_id(self, label: str, object_id: int, patterns: List[str], root_directory: Optional[Path]) -> Path:
        if root_directory is None:
            root_directory = self._lean_config_manager.get_cli_root_directory()

        for pattern in patterns:
            for directory in root_directory.rglob(pattern):
                if not directory.is_dir():
                    continue

                config = self.get_output_config(directory)
                if config.get("id", None) == object_id:
                    return directory

        raise ValueError(f"{label} with id '{object_id}' does not exist")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import os
import platform
import site
import sys
import time
from pathlib import Path


# Docker's pywin32 dependency on Windows is a common source of issues
# In a lot of cases you'd have to manually run pywin32's post-install script as admin after pip installing the library
# This is a hassle, so the CLI attempts to automate this step when it's necessary
# Additionally, we can also fix the issues for some users by updating os.environ["PATH"]
# If this works we use this fix instead as it removes the need to request admin permissions
# This code must run before the Docker package is imported anywhere in the code

def _is_win32_available() -> bool:
    try:
        # Try the win32 APIs used by https://github.com/docker/docker-py/blob/master/docker/transport/npipesocket.py
        import win32file
        import win32pipe
        return True
    except:
        return False


def _ensure_win32_available() -> None:
    if _is_win32_available():
        return

    possible_paths = sys.path + [sys.prefix] + site.getsitepackages() + [site.getusersitepackages()]
    possible_paths = [Path(p) for p in possible_paths]
    possible_directories = set(p for p in possible_paths if p.is_dir())

    for directory in possible_directories:
        target_directory = directory / "pywin32_system32"
        if not target_directory.is_dir():
            continue

        os.environ["PATH"] += ";" + str(target_directory)

        if _is_win32_available():
            return

    for directory in possible_directories:
        target_file = directory / "Scripts" / "pywin32_postinstall.py"
        if not target_file.is_file():
            continue

        print(f"Running pywin32's post-install script at {target_file}")
        ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, f'"{target_file}" -install', None, 1)

        # ShellExecuteW returns immediately after the UAC dialog, we wait a second to give the script some time to run
        time.sleep(1)

        if _is_win32_available():
            return

    if any("AppData\\Local\\Packages\\PythonSoftwareFoundation.Python" in p for p in sys.path):
        print("It looks like you're using the Python distribution from the Microsoft Store")
        print("This distribution is not supported by the CLI, we recommend using the Anaconda distribution instead")
        print(
            "See https://www.lean.io/docs/v2/lean-cli/installation/installing-pip#02-Install-on-Windows for more information")
        sys.exit(1)

    print("pywin32 has not been installed completely, which may lead to errors")
    print("You can fix this issue by running pywin32's post-install script")
    print(f"Run the following command in an elevated terminal from your Python environment's Scripts directory:")
    print("python pywin32_postinstall.py -install")


if platform.system() == "Windows":
    _ensure_win32_available()

import traceback
from io import StringIO

import click
import requests
from pydantic import ValidationError

from lean.commands import lean
from lean.container import container
from lean.models.errors import MoreInfoError


def main() -> None:
    """This function is the entrypoint when running a Lean command in a terminal."""
    try:
        lean.main(standalone_mode=False)

        temp_manager = container.temp_manager()
        if temp_manager.delete_temporary_directories_when_done:
            temp_manager.delete_temporary_directories()
    except Exception as exception:
        logger = container.logger()
        logger.debug(traceback.format_exc().strip())

        if isinstance(exception, ValidationError) and hasattr(exception, "input_value"):
            logger.debug("Value that failed validation:")
            logger.debug(exception.input_value)
            logger.error(f"Error: {exception}")
        elif isinstance(exception, MoreInfoError):
            logger.error(f"Error: {exception}")
            logger.error(f"Visit {exception.link} for more information")
        elif isinstance(exception, click.UsageError):
            io = StringIO()
            exception.show(file=io)

            exception_str = io.getvalue().strip()
            exception_str = exception_str.replace("Try 'lean", "\nTry 'lean")
            exception_str = exception_str.replace("for help.",
                                                  "for help or go to the following url for a list of common errors:\nhttps://www.lean.io/docs/v2/lean-cli/key-concepts/troubleshooting#02-Common-Errors")

            container.update_manager().warn_if_cli_outdated(force=True)

            logger.info(exception_str)
        elif isinstance(exception, click.Abort):
            logger.info("Aborted!")
        elif isinstance(exception, requests.exceptions.ConnectionError):
            logger.error(f"Error: {exception}")
            logger.error("It looks like you don't have an internet connection, please check your system settings")
        else:
            logger.error(f"Error: {exception}")

        temp_manager = container.temp_manager()
        if temp_manager.delete_temporary_directories_when_done:
            temp_manager.delete_temporary_directories()

        sys.exit(1)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
from rich import box
from rich.table import Table

from lean.click import LeanCommand
from lean.container import container


@click.command(cls=LeanCommand)
def list() -> None:
    """List the configurable options and their current values."""
    table = Table(box=box.SQUARE)
    table.add_column("Key", overflow="fold")
    table.add_column("Value", overflow="fold")
    table.add_column("Location", overflow="fold")
    table.add_column("Description", overflow="fold")

    for option in container.cli_config_manager().all_options:
        value = option.get_value(default="<not set>")

        # Mask values of sensitive options
        if value != "<not set>" and option.is_sensitive:
            value = "*" * 12 + value[-3:] if len(value) >= 5 else "*" * 15

        table.add_row(option.key,
                      value,
                      str(option.location),
                      option.description)

    logger = container.logger()
    logger.info(table)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import traceback
from hashlib import sha256
from time import time
from typing import Any, Dict
from urllib.parse import urljoin

import requests

import lean
from lean.components.api.account_client import AccountClient
from lean.components.api.backtest_client import BacktestClient
from lean.components.api.compile_client import CompileClient
from lean.components.api.data_client import DataClient
from lean.components.api.file_client import FileClient
from lean.components.api.lean_client import LeanClient
from lean.components.api.live_client import LiveClient
from lean.components.api.market_client import MarketClient
from lean.components.api.module_client import ModuleClient
from lean.components.api.node_client import NodeClient
from lean.components.api.optimization_client import OptimizationClient
from lean.components.api.organization_client import OrganizationClient
from lean.components.api.project_client import ProjectClient
from lean.components.api.service_client import ServiceClient
from lean.components.api.user_client import UserClient
from lean.components.util.http_client import HTTPClient
from lean.components.util.logger import Logger
from lean.constants import API_BASE_URL
from lean.models.errors import AuthenticationError, RequestFailedError


class APIClient:
    """The APIClient class manages communication with the QuantConnect API."""

    def __init__(self, logger: Logger, http_client: HTTPClient, user_id: str, api_token: str) -> None:
        """Creates a new APIClient instance.

        :param logger: the logger to use to print debug messages to
        :param http_client: the HTTP client to make HTTP requests with
        :param user_id: the QuantConnect user id to use when sending authenticated requests
        :param api_token: the QuantConnect API token to use when sending authenticated requests
        """
        self._logger = logger
        self._http_client = http_client
        self._user_id = user_id
        self._api_token = api_token

        # Create the clients containing the methods to send requests to the various API endpoints
        self.accounts = AccountClient(self)
        self.backtests = BacktestClient(self)
        self.compiles = CompileClient(self)
        self.data = DataClient(self, http_client)
        self.files = FileClient(self)
        self.live = LiveClient(self)
        self.market = MarketClient(self)
        self.modules = ModuleClient(self)
        self.nodes = NodeClient(self)
        self.optimizations = OptimizationClient(self)
        self.organizations = OrganizationClient(self)
        self.projects = ProjectClient(self)
        self.services = ServiceClient(self)
        self.users = UserClient(self)
        self.lean = LeanClient(self)

    def get(self, endpoint: str, parameters: Dict[str, Any] = {}) -> Any:
        """Makes an authenticated GET request to the given endpoint with the given parameters.

        Raises an error if the request fails or if the current credentials are invalid.

        :param endpoint: the API endpoint to send the request to
        :param parameters: the parameters to attach to the url
        :return: the parsed response of the request
        """
        return self._request("get", endpoint, {"params": parameters})

    def post(self, endpoint: str, data: Dict[str, Any] = {}, data_as_json: bool = True) -> Any:
        """Makes an authenticated POST request to the given endpoint with the given data.

        Raises an error if the request fails or if the current credentials are invalid.

        :param endpoint: the API endpoint to send the request to
        :param data: the data to send in the body of the request
        :param data_as_json: True if data needs to be sent as JSON, False if data needs to be sent as form data
        :return: the parsed response of the request
        """
        options = {"json": data} if data_as_json else {"data": data}
        return self._request("post", endpoint, options)

    def is_authenticated(self) -> bool:
        """Checks whether the current credentials are valid.

        :return: True if the current credentials are valid, False if not
        """
        try:
            self.get("authenticate")
            return True
        except (RequestFailedError, AuthenticationError):
            self._logger.debug(traceback.format_exc().strip())
            return False

    def _request(self, method: str, endpoint: str, options: Dict[str, Any] = {}, retry_http_5xx: bool = True) -> Any:
        """Makes an authenticated request to the given endpoint.

        :param method: the HTTP method to use for the request
        :param endpoint: the API endpoint to send the request to
        :param options: additional options to pass on to requests.request()
        :param retry_http_5xx: True if the request should be retried on an HTTP 5xx response, False if not
        :return: the parsed response of the request
        """
        full_url = urljoin(API_BASE_URL, endpoint)

        # Create the hash which is used to authenticate the user to the API
        timestamp = str(int(time()))
        password = sha256(f"{self._api_token}:{timestamp}".encode("utf-8")).hexdigest()

        headers = {
            "Timestamp": timestamp
        }

        version = lean.__version__
        if lean.__version__ == 'dev':
            version = 99999999
        headers["User-Agent"] = f"Lean CLI {version}"

        response = self._http_client.request(method,
                                             full_url,
                                             headers=headers,
                                             auth=(self._user_id, password),
                                             raise_for_status=False,
                                             **options)

        if 500 <= response.status_code < 600 and retry_http_5xx:
            return self._request(method, endpoint, options, False)

        if response.status_code == 500:
            raise AuthenticationError()

        if response.status_code < 200 or response.status_code >= 300:
            raise RequestFailedError(response)

        return self._parse_response(response)

    def _parse_response(self, response: requests.Response) -> Any:
        """Parses the data in a response.

        Raises an error if the data in the response indicates something went wrong.

        :param response: the response of the request
        :return: the data in the response
        """
        data = response.json()

        if data["success"]:
            return data

        self._http_client.log_unsuccessful_response(response)

        if "errors" in data and len(data["errors"]) > 0:
            if data["errors"][0].startswith("Hash doesn't match."):
                raise AuthenticationError()
            if data["errors"][0].startswith('UserID not valid'):
                data["errors"].append('Please login to your account. https://www.quantconnect.com/docs/v2/lean-cli/api-reference/lean-login')
            raise RequestFailedError(response, "\n".join(data["errors"]))

        if "messages" in data and len(data["messages"]) > 0:
            raise RequestFailedError(response, "\n".join(data["messages"]))

        if "Message" in data:
            raise RequestFailedError(response, data["Message"])

        raise RequestFailedError(response)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from lean.click import LeanCommand
from lean.container import container
from lean.models.api import QCLiveAlgorithmStatus
from lean.models.brokerages.cloud import all_cloud_brokerages, PaperTradingBrokerage


@click.command(cls=LeanCommand)
@click.argument("project", type=str)
def status(project: str) -> None:
    """Show the live trading status of a project in the cloud.

    PROJECT must be the name or the id of the project to show the status for.
    """
    logger = container.logger()
    api_client = container.api_client()

    cloud_project_manager = container.cloud_project_manager()
    cloud_project = cloud_project_manager.get_cloud_project(project, False)

    live_algorithm = next((d for d in api_client.live.get_all() if d.projectId == cloud_project.projectId), None)

    logger.info(f"Project id: {cloud_project.projectId}")
    logger.info(f"Project name: {cloud_project.name}")
    logger.info(f"Project url: {cloud_project.get_url()}")

    if live_algorithm is None:
        logger.info("Live status: Not deployed")
        return

    live_status = {
        QCLiveAlgorithmStatus.DeployError: "Deploy error",
        QCLiveAlgorithmStatus.InQueue: "In queue",
        QCLiveAlgorithmStatus.RuntimeError: "Runtime error",
        QCLiveAlgorithmStatus.LoggingIn: "Logging in"
    }.get(live_algorithm.status, live_algorithm.status.value)

    brokerage_name = next((b.get_name() for b in all_cloud_brokerages if b.get_id() == live_algorithm.brokerage),
                          live_algorithm.brokerage)

    if brokerage_name == "PaperBrokerage":
        brokerage_name = PaperTradingBrokerage.get_name()

    logger.info(f"Live status: {live_status}")
    logger.info(f"Live id: {live_algorithm.deployId}")
    logger.info(f"Live url: {live_algorithm.get_url()}")
    logger.info(f"Brokerage: {brokerage_name}")
    logger.info(f"Launched: {live_algorithm.launched.strftime('%Y-%m-%d %H:%M:%S')} UTC")

    if live_algorithm.stopped is not None:
        logger.info(f"Stopped: {live_algorithm.stopped.strftime('%Y-%m-%d %H:%M:%S')} UTC")

    if live_algorithm.error != "":
        logger.info("Error:")
        logger.info(live_algorithm.error)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from lean.click import LeanCommand
from lean.container import container


@click.command(cls=LeanCommand)
@click.argument("key", type=str)
def unset(key: str) -> None:
    """Unset a configurable option.

    Run `lean config list` to show all available options.
    """
    cli_config_manager = container.cli_config_manager()

    option = cli_config_manager.get_option_by_key(key)
    option.unset()

    logger = container.logger()
    logger.info(f"Successfully unset '{key}'")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lxml import etree


class XMLManager:
    """The PathManager class provides utilities for working with XML."""

    def __init__(self) -> None:
        """Creates a new XMLManager instance."""
        self._xml_parser = etree.XMLParser(remove_blank_text=True, encoding="utf-8")

    def parse(self, xml_string: str) -> etree.Element:
        """Parses an XML string to an XML element.

        :param xml_string: the string to parse
        :return: the parsed XML element
        """
        return etree.fromstring(xml_string.encode("utf-8"), parser=self._xml_parser)

    def to_string(self, root: etree.Element) -> str:
        """Turns an XML element into a pretty string.

        :param root: the XML element to turn into a string
        :return: the XML element as a pretty string using 4 spaces as indentation
        """
        etree.indent(root, " " * 4)
        return etree.tostring(root, encoding="utf-8", method="xml", pretty_print=True).decode("utf-8")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import List, Dict, Any

from pydantic import validator

from lean.models.pydantic import WrappedBaseModel


class MarketHoursSegment(WrappedBaseModel):
    start: str
    end: str
    state: str


class MarketHoursDatabaseEntry(WrappedBaseModel):
    dataTimeZone: str
    exchangeTimeZone: str
    monday: List[MarketHoursSegment] = []
    tuesday: List[MarketHoursSegment] = []
    wednesday: List[MarketHoursSegment] = []
    thursday: List[MarketHoursSegment] = []
    friday: List[MarketHoursSegment] = []
    saturday: List[MarketHoursSegment] = []
    sunday: List[MarketHoursSegment] = []
    holidays: List[datetime] = []
    earlyCloses: Dict[str, str] = {}
    lateOpens: Dict[str, str] = {}

    @validator("holidays", pre=True)
    def parse_holidays(cls, value: Any) -> Any:
        if isinstance(value, list):
            return [datetime.strptime(x, "%m/%d/%Y") for x in value]
        return value



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import shutil
import subprocess
from pathlib import Path

from lxml import etree

from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.util.logger import Logger
from lean.components.util.path_manager import PathManager
from lean.components.util.project_manager import ProjectManager
from lean.components.util.xml_manager import XMLManager
from lean.models.utils import LeanLibraryReference


class LibraryManager:
    """The LibraryManager class provides utilities for handling a libraries."""

    def __init__(self,
                 logger: Logger,
                 project_manager: ProjectManager,
                 project_config_manager: ProjectConfigManager,
                 lean_config_manager: LeanConfigManager,
                 path_manager: PathManager,
                 xml_manager: XMLManager) -> None:
        """Creates a new LibraryManager instance.

        :param logger: the logger to use to log messages with
        :param project_manager: the ProjectManager to use when requesting project csproj file
        :param project_config_manager: the ProjectConfigManager to use to get project's config
        :param lean_config_manager: the LeanConfigManager to get the CLI root directory from
        :param path_manager: the path manager to use to handle library paths
        :param xml_manager: the XMLManager to use when working with XML
        """
        self._logger = logger
        self._project_manager = project_manager
        self._project_config_manager = project_config_manager
        self._lean_config_manager = lean_config_manager
        self._path_manager = path_manager
        self._xml_manager = xml_manager

    def is_lean_library(self, path: Path) -> bool:
        """Checks whether the library name is a Lean library path

        :param path: path to check whether it is a Lean library
        :return: true if the path is a Lean library path
        """
        if not self._path_manager.is_path_valid(path):
            return False

        relative_path = self._path_manager.get_relative_path(path, self._lean_config_manager.get_cli_root_directory())
        path_parts = relative_path.parts
        library_config = self._project_config_manager.get_project_config(path)
        library_language = library_config.get("algorithm-language", None)

        return (
            len(path_parts) > 0 and
            path_parts[0] == "Library" and
            relative_path.is_dir() and
            library_language is not None
        )

    def get_csharp_lean_library_path_for_csproj_file(self, project_dir: Path, library_dir: Path) -> str:
        """Gets the library path to be used for the project's .csproj file.

        The returned path is relative to the project directory so auto complete can be provided.

        :param project_dir: The path to the project directory
        :param library_dir: The path to the library directory
        :return The path to be used for the project's .csproj file
        """
        cli_root_dir = self._lean_config_manager.get_cli_root_directory()
        project_dir_relative_to_cli = self._path_manager.get_relative_path(project_dir, cli_root_dir)
        library_dir_relative_to_cli = self._path_manager.get_relative_path(library_dir, cli_root_dir)
        library_csproj_file = self._project_manager.get_csproj_file_path(library_dir_relative_to_cli)

        if library_csproj_file is None:
            return None

        return ("../" * len(project_dir_relative_to_cli.parts) / library_csproj_file).as_posix()

    def get_library_path_for_project_config_file(self, library_dir: Path) -> str:
        """Gets the library path to be used for the project's config.json file.

        :param library_dir: The path to the library directory
        :return The path to be used for the project's config.json file
        """
        lean_cli_root_dir = self._lean_config_manager.get_cli_root_directory()
        return self._path_manager.get_relative_path(library_dir, lean_cli_root_dir).as_posix()

    def add_lean_library_reference_to_project(self, project_dir: Path, library_dir: Path) -> bool:
        """Adds a Lean CLI library reference to a project.

        Adds the library path to the project's config.json

        :param project_dir: the path to the project directory
        :param library_dir: the path to the library directory
        :return: True if the library has already been added to the project, that is, if the reference already exists in
                 the project's config.json file. False if the library was added successfully.
        """
        project_config = self._project_config_manager.get_project_config(project_dir)
        project_libraries = project_config.get("libraries", [])
        library_relative_path = Path(self.get_library_path_for_project_config_file(library_dir))

        if any(LeanLibraryReference(**library).path == library_relative_path for library in project_libraries):
            return True

        library_config = self._project_config_manager.get_project_config(library_dir)
        library_libraries = library_config.get("libraries", [])
        project_relative_path = Path(self.get_library_path_for_project_config_file(project_dir))

        if any(LeanLibraryReference(**library).path == project_relative_path for library in library_libraries):
            raise RuntimeError("Circular dependency detected between "
                               f"{project_relative_path} and {library_relative_path}")

        project_libraries.append(json.loads(LeanLibraryReference(
            name=library_dir.name,
            path=library_relative_path
        ).json()))
        project_config.set("libraries", project_libraries)

        return False

    def remove_lean_library_reference_from_project(self, project_dir: Path, library_dir: Path) -> None:
        """Removed a Lean CLI library reference from a project.

        Removes the library path from the project's config.json

        :param project_dir: the path to the project directory
        :param library_dir: the path to the C# library directory
        """
        library_relative_path = Path(self.get_library_path_for_project_config_file(library_dir))
        project_config = self._project_config_manager.get_project_config(project_dir)
        libraries = project_config.get("libraries", [])
        libraries = [library for library in libraries if LeanLibraryReference(**library).path != library_relative_path]
        project_config.set("libraries", libraries)

    def add_lean_library_to_csharp_project(self, project_dir: Path, library_dir: Path, no_local: bool) -> None:
        """Adds a Lean CLI C# library to the project's csproj file.

        It will add a reference to the library and restore the project if dotnet is the user's PATH.

        It will raise a RuntimeError if the library is a C# library and does not have a .csproj file.

        :param project_dir: Path to the project directory
        :param library_dir: Path to the library directory
        :param no_local: Whether restoring the packages locally must be skipped
        """
        already_added = self.add_lean_library_reference_to_project(project_dir, library_dir)
        # If the library was already referenced by the project,
        # do not proceed further with adding it to the .csproj file
        if already_added:
            self._logger.info(f"Library {library_dir.name} has already been added to the project {project_dir.name}")
            return

        library_config = self._project_config_manager.get_project_config(library_dir)
        library_language = library_config.get("algorithm-language", None)

        # Python library references are not added to .csproj file
        if library_language == 'Python':
            return

        library_csproj_file_path = self.get_csharp_lean_library_path_for_csproj_file(project_dir, library_dir)

        if library_csproj_file_path is None:
            raise RuntimeError(f"C# library {library_dir.name} does not contain a .csproj file")

        project_csproj_file = self._project_manager.get_csproj_file_path(project_dir)

        original_csproj_content = project_csproj_file.read_text(encoding="utf-8")
        self._add_csharp_project_to_csproj(project_csproj_file, library_csproj_file_path)
        self._project_manager.try_restore_csharp_project(project_csproj_file, original_csproj_content, no_local)

    def add_lean_library_to_python_project(self, project_dir: Path, library_dir: Path) -> None:
        """Adds a Lean CLI Python library to a Python project.

        :param project_dir: the path to the project directory
        :param library_dir: the path to the library directory
        """
        self.add_lean_library_reference_to_project(project_dir, library_dir)

    def add_lean_library_to_project(self, project_dir: Path, library_dir: Path, no_local: bool) -> None:
        """Adds a Lean CLI library to a project.

        :param project_dir: the path to the project directory
        :param library_dir: the path to the library directory
        :param no_local: whether restoring the packages locally must be skipped
        """
        project_language = self._get_project_language(project_dir)

        if project_language == "CSharp":
            self.add_lean_library_to_csharp_project(project_dir, library_dir, no_local)
        else:
            self.add_lean_library_to_python_project(project_dir, library_dir)

    def remove_lean_library_from_csharp_project(self, project_dir: Path, library_dir: Path, no_local: bool) -> None:
        """Removes a Lean CLI library from a C# project.

        Removes the library from the project's .csproj file,
        and restores the project if dotnet is on the user's PATH and no_local is False.

        :param project_dir: Path to the project directory
        :param library_dir: Path to the library directory
        :param no_local: Whether restoring the packages locally must be skipped
        """
        self.remove_lean_library_reference_from_project(project_dir, library_dir)

        library_config = self._project_config_manager.get_project_config(library_dir)
        library_language = library_config.get("algorithm-language", None)

        # Python library references are not added to .csproj file, so no need to remove
        if library_language == 'Python':
            return

        self._remove_project_reference_from_csharp_project(project_dir, library_dir, no_local)

    def remove_lean_library_from_python_project(self, project_dir: Path, library_dir: Path) -> None:
        """Removes a Lean CLI library from a Python project.

        :param project_dir: Path to the project directory
        :param library_dir: Path to the library directory
        """
        self.remove_lean_library_reference_from_project(project_dir, library_dir)

    def remove_lean_library_from_project(self, project_dir: Path, library_dir: Path, no_local: bool) -> None:
        """Removes a Lean CLI library from a project.

        Removes the library from the project's .csproj file,
        and restores the project if dotnet is on the user's PATH and no_local is False.

        :param project_dir: Path to the project directory
        :param library_dir: Path to the library directory
        :param no_local: Whether restoring the packages locally must be skipped
        """
        project_language = self._get_project_language(project_dir)

        if project_language == "CSharp":
            self.remove_lean_library_from_csharp_project(project_dir, library_dir, no_local)
        else:
            self.remove_lean_library_from_python_project(project_dir, library_dir)

    def _get_project_language(self, project_dir: Path) -> str:
        """Gets a project language.

        :param project_dir: Path to the project directory
        :return The project language
        """
        project_config = self._project_config_manager.get_project_config(project_dir)
        return project_config.get("algorithm-language", None)

    def _add_csharp_project_to_csproj(self, csproj_file: Path, library_csproj_path: str) -> None:
        """Adds a Lean CLI library project reference to a .csproj file.

        :param csproj_file: the path to the .csproj file
        :param library_csproj_path: the path to the library's .csproj file
        """
        csproj_tree = self._xml_manager.parse(csproj_file.read_text(encoding="utf-8"))

        existing_project_reference = csproj_tree.find(f".//ProjectReference[@Include='{library_csproj_path}']")
        if existing_project_reference is None:
            last_item_group = csproj_tree.find(".//ItemGroup[last()]")
            if last_item_group is None:
                last_item_group = etree.SubElement(csproj_tree.find(".//Project"), "ItemGroup")

            last_item_group.append(etree.fromstring(f'<ProjectReference Include="{library_csproj_path}" />'))

        csproj_file.write_text(self._xml_manager.to_string(csproj_tree), encoding="utf-8")

    def _remove_project_reference_from_csharp_project(self, project_dir: Path, name: Path, no_local: bool) -> None:
        """Removes a Lean CLI C# library from a C# project.

        Removes the library from the project's .csproj file,
        and restores the project if dotnet is on the user's PATH and no_local is False.

        :param project_dir: the path to the project directory
        :param name: the path to the Lean CLI library to remove
        :param no_local: Whether restoring the packages locally must be skipped
        """
        csproj_file = self._project_manager.get_csproj_file_path(project_dir)
        self._logger.info(f"Removing {name} from '{self._path_manager.get_relative_path(csproj_file)}'")

        csproj_tree = self._xml_manager.parse(csproj_file.read_text(encoding="utf-8"))

        library_reference = self.get_csharp_lean_library_path_for_csproj_file(project_dir, name)

        for package_reference in csproj_tree.findall('.//ProjectReference'):
            if package_reference.get("Include", "") == library_reference:
                package_reference.getparent().remove(package_reference)

        csproj_file.write_text(self._xml_manager.to_string(csproj_tree), encoding="utf-8")

        if not no_local and shutil.which("dotnet") is not None:
            self._logger.info(f"Restoring packages in '{self._path_manager.get_relative_path(project_dir)}'")

            process = subprocess.run(["dotnet", "restore", str(csproj_file)], cwd=project_dir)

            if process.returncode != 0:
                raise RuntimeError(
                    "Something went wrong while restoring packages, see the logs above for more information")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import itertools
import re
from time import sleep
import webbrowser
from collections import OrderedDict
from typing import Iterable, List, Optional

import click
from rich import box
from rich.table import Table

from lean.click import LeanCommand, ensure_options
from lean.container import container
from lean.models.api import QCDataInformation, QCDataVendor, QCFullOrganization, QCDatasetDelivery
from lean.models.data import Dataset, DataFile, Product
from lean.models.logger import Option

_data_information: Optional[QCDataInformation] = None
_presigned_terms="""
Data Terms of Use has been signed previously.
Find full agreement at: {link}
==========================================================================
CLI API Access Agreement: On {signed_time} You Agreed:
- Display or distribution of data obtained through CLI API Access is not permitted. 
- Data and Third Party Data obtained via CLI API Access can only be used for individual or internal employee's use.
- Data is provided in LEAN format can not be manipulated for transmission or use in other applications.
- QuantConnect is not liable for the quality of data received and is not responsible for trading losses.
==========================================================================
"""

def _get_data_information(organization: QCFullOrganization) -> QCDataInformation:
    """Retrieves the datasources and prices information.

    :param organization: the organization to get the information for
    :return: the datasources and prices information
    """
    global _data_information

    if _data_information is None:
        _data_information = container.api_client().data.get_info(organization.id)

    return _data_information


def _map_data_files_to_vendors(organization: QCFullOrganization, data_files: Iterable[str]) -> List[DataFile]:
    """Maps a list of data files to the available data vendors.

    Uses the API to get the latest price information.
    Raises an error if there is no vendor that sells the data of a file in the given list.

    :param organization: the organization to use the price information of
    :param data_files: the data files to map to the available vendors
    :return: the list of data files containing the file and vendor for each file
    """
    data_information = _get_data_information(organization)

    last_vendor: Optional[QCDataVendor] = None
    mapped_files = []

    for file in data_files:
        if last_vendor is not None and last_vendor.regex.search(file):
            mapped_files.append(DataFile(file=file, vendor=last_vendor))
            continue

        last_vendor = None

        for vendor in data_information.prices:
            if vendor.price is None:
                continue

            if vendor.regex.search(file):
                mapped_files.append(DataFile(file=file, vendor=vendor))
                last_vendor = vendor
                break

        if last_vendor is None:
            raise RuntimeError(f"There is no data vendor that sells '{file}'")

    return mapped_files


def _get_data_files(organization: QCFullOrganization, products: List[Product]) -> List[DataFile]:
    """Returns the unique data files of a list of products mapped to their vendor.

    :param organization: the organization to use the price information of
    :param products: the list of products to get the data files from
    :return: the list of unique data files containing the file and vendor for each file for each product
    """
    unique_data_files = sorted(list(set(itertools.chain(*[product.get_data_files() for product in products]))))
    return _map_data_files_to_vendors(organization, unique_data_files)


def _display_products(organization: QCFullOrganization, products: List[Product]) -> None:
    """Previews a list of products in pretty tables.

    :param organization: the organization the user selected
    :param products: the products to display
    """
    logger = container.logger()
    table = Table(box=box.SQUARE)

    for column in ["Dataset", "Vendor", "Details", "File count", "Price"]:
        table.add_column(column, overflow="fold")

    summed_price = 0

    for product in products:
        details = []
        for option_id, result in product.option_results.items():
            option = next(o for o in product.dataset.options if o.id == option_id)
            if result is not None:
                label = option.label

                if isinstance(result.value, list):
                    if len(result.value) > 1:
                        label = label.replace("(s)", "s")
                    else:
                        label = label.replace("(s)", "")

                details.append(f"{label}: {result.label}")

        if len(details) == 0:
            details.append("-")

        mapped_files = _map_data_files_to_vendors(organization, product.get_data_files())
        price = sum(data_file.vendor.price for data_file in mapped_files)
        summed_price += price

        table.add_row(product.dataset.name,
                      product.dataset.vendor,
                      "\n".join(details),
                      f"{len(mapped_files):,.0f}",
                      f"{price:,.0f} QCC")

    logger.info(table)

    all_data_files = _get_data_files(organization, products)
    total_price = sum(data_file.vendor.price for data_file in all_data_files)

    if total_price != summed_price:
        logger.warn("The total price is less than the sum of all separate prices because there is overlapping data")

    logger.info(f"Total price: {total_price:,.0f} QCC")
    logger.info(f"Organization balance: {organization.credit.balance:,.0f} QCC")


def _select_organization() -> QCFullOrganization:
    """Asks the user for the organization that should be used.

    :return: the selected organization
    """
    api_client = container.api_client()

    organizations = api_client.organizations.get_all()
    options = [Option(id=organization.id, label=organization.name) for organization in organizations]

    logger = container.logger()
    organization_id = logger.prompt_list("Select the organization to purchase and download data with", options)

    return api_client.organizations.get(organization_id)


def _select_products_interactive(organization: QCFullOrganization, datasets: List[Dataset]) -> List[Product]:
    """Asks the user for the products that should be purchased and downloaded.

    :param organization: the organization that will be charged
    :param datasets: the available datasets
    :return: the list of products selected by the user
    """
    products = []
    logger = container.logger()

    category_options = {}
    for dataset in datasets:
        for category in dataset.categories:
            if category in category_options:
                continue

            dataset_count = len(list(dataset for dataset in datasets if category in dataset.categories))
            category_options[category] = Option(
                id=category,
                label=f"{category} ({dataset_count} dataset{'s' if dataset_count > 1 else ''})"
            )

    category_options = sorted(category_options.values(), key=lambda opt: opt.label)

    while True:
        category = logger.prompt_list("Select a category", category_options)

        available_datasets = sorted((d for d in datasets if category in d.categories), key=lambda d: d.name)
        dataset: Dataset = logger.prompt_list("Select a dataset",
                                              [Option(id=d, label=d.name) for d in available_datasets])

        if dataset.requires_security_master and not organization.has_security_master_subscription():
            logger.warn("\n".join([
                f"Your organization needs to have an active Security Master subscription to download data from the '{dataset.name}' dataset",
                f"You can add the subscription at https://www.quantconnect.com/datasets/quantconnect-security-master/pricing"
            ]))
            continue

        option_results = OrderedDict()
        for option in dataset.options:
            if option.condition is None or option.condition.check(option_results):
                option_results[option.id] = option.configure_interactive()

        products.append(Product(dataset=dataset, option_results=option_results))

        logger.info("Selected data:")
        _display_products(organization, products)

        if not click.confirm("Do you want to download more data?"):
            break

    return products


def _confirm_organization_balance(organization: QCFullOrganization, products: List[Product]) -> None:
    """Checks whether the selected organization has enough QCC to download all selected data.

    Raises an error if the organization does not have enough QCC.

    :param organization: the organization that the user selected
    :param products: the list of products selected by the user
    """
    all_data_files = _get_data_files(organization, products)
    total_price = sum(data_file.vendor.price for data_file in all_data_files)

    if total_price > organization.credit.balance:
        raise RuntimeError("\n".join([
            "The total price exceeds your organization's QCC balance",
            "You can purchase QCC by clicking \"Purchase Credit\" on your organization's home page:",
            f"https://www.quantconnect.com/organization/{organization.id}"
        ]))


def _verify_accept_agreement(organization: QCFullOrganization, open_browser: bool) -> None:
    """ Verifies that the user has accepted the agreement.
    If they haven't, asks the user to accept the CLI API Access and Data Agreement.
    If they have, reminds them of the agreement and moves on.

    The API will enforce signing the agreement at the end of the day but this is how we keep it in the process of the CLI

    :param organization: the organization that the user selected
    :param open_browser: whether the CLI should automatically open the agreement in the browser
    """
    logger = container.logger()
    api_client = container.api_client()

    info = api_client.data.get_info(organization.id)

    # Is signed
    if organization.data.current:
        logger.info(_presigned_terms.format(link=info.agreement, signed_time=datetime.fromtimestamp(organization.data.signedTime)))
        sleep(1)
    else:
        if open_browser:
            webbrowser.open(info.agreement)

        logger.info(f"Go to the following url to accept the CLI API Access and Data Agreement:")
        logger.info(info.agreement)
        logger.info("Waiting until the CLI API Access and Data Agreement has been accepted...")

        container.task_manager().poll(
            make_request=lambda: api_client.organizations.get(organization.id),
            is_done=lambda data: data.data.current != False
        )


def _confirm_payment(organization: QCFullOrganization, products: List[Product]) -> None:
    """Processes payment for the selected products.

    An abort error is raised if the user decides to cancel.

    :param organization: the organization that will be charged
    :param products: the list of products selected by the user
    """
    all_data_files = _get_data_files(organization, products)
    total_price = sum(data_file.vendor.price for data_file in all_data_files)

    organization_qcc = organization.credit.balance

    logger = container.logger()
    logger.info(f"You will be charged {total_price:,.0f} QCC from your organization's QCC balance")
    logger.info(
        f"After downloading all files your organization will have {organization_qcc - total_price:,.0f} QCC left")

    click.confirm("Continue?", abort=True)


def _get_organization_by_name_or_id(user_input: str) -> QCFullOrganization:
    """Finds an organization by name or id.

    Raises an error if no organization with a matching name or id can be found.

    :param user_input: the input given by the user
    :return: the first organization with the given name or id
    """
    api_client = container.api_client()

    if re.match("^[a-f0-9]{32}$", user_input) is not None:
        try:
            return api_client.organizations.get(user_input)
        except:
            pass

    all_organizations = api_client.organizations.get_all()
    selected_organization = next((o for o in all_organizations if o.id == user_input or o.name == user_input), None)

    if selected_organization is None:
        raise RuntimeError(f"You are not a member of an organization with name or id '{user_input}'")

    return api_client.organizations.get(selected_organization.id)


def _select_products_non_interactive(organization: QCFullOrganization,
                                     datasets: List[Dataset],
                                     ctx: click.Context) -> List[Product]:
    """Asks the user for the products that should be purchased and downloaded.

    :param organization: the organization that will be charged
    :param datasets: the available datasets
    :param ctx: the click context of the invocation
    :return: the list of products selected by the user
    """
    dataset = next((d for d in datasets if d.name.lower() == ctx.params["dataset"].lower()), None)
    if dataset is None:
        raise RuntimeError(f"There is no dataset named '{ctx.params['dataset']}'")

    if dataset.requires_security_master and not organization.has_security_master_subscription():
        raise RuntimeError("\n".join([
            f"Your organization needs to have an active Security Master subscription to download data from the '{dataset.name}' dataset",
            f"You can add the subscription at https://www.quantconnect.com/datasets/quantconnect-security-master/pricing"
        ]))

    option_results = OrderedDict()
    invalid_options = []
    missing_options = []

    for option in dataset.options:
        if option.condition is not None and not option.condition.check(option_results):
            continue

        user_input = ctx.params.get(option.id, None)

        if user_input is None:
            missing_options.append(f"--{option.id} <{option.get_placeholder()}>: {option.description}")
        else:
            try:
                option_results[option.id] = option.configure_non_interactive(user_input)
            except ValueError as error:
                invalid_options.append(f"--{option.id}: {error}")

    if len(invalid_options) > 0 or len(missing_options) > 0:
        blocks = []

        for label, lines in [["Invalid option", invalid_options], ["Missing option", missing_options]]:
            if len(lines) > 0:
                joined_lines = "\n".join(lines)
                blocks.append(f"{label}{'s' if len(lines) > 1 else ''}:\n{joined_lines}")

        raise RuntimeError("\n\n".join(blocks))

    products = [Product(dataset=dataset, option_results=option_results)]

    container.logger().info("Data that will be purchased and downloaded:")
    _display_products(organization, products)

    return products


def _get_available_datasets(organization: QCFullOrganization) -> List[Dataset]:
    """Retrieves the available datasets.

    :param organization: the organization that will be charged
    :return: the datasets which data can be downloaded from
    """
    cloud_datasets = container.api_client().market.list_datasets()
    data_information = _get_data_information(organization)

    available_datasets = []
    for cloud_dataset in cloud_datasets:
        if cloud_dataset.delivery == QCDatasetDelivery.CloudOnly:
            continue

        datasource = data_information.datasources.get(str(cloud_dataset.id), None)
        if datasource is None or isinstance(datasource, list):
            if cloud_dataset.name != "Template Data Source Product":
                name = cloud_dataset.name.strip()
                vendor = cloud_dataset.vendorName.strip()
                container.logger().debug(
                    f"There is no datasources entry for {name} by {vendor} (id {cloud_dataset.id})")
            continue

        available_datasets.append(Dataset(name=cloud_dataset.name.strip(),
                                          vendor=cloud_dataset.vendorName.strip(),
                                          categories=[tag.name.strip() for tag in cloud_dataset.tags],
                                          options=datasource["options"],
                                          paths=datasource["paths"],
                                          requires_security_master=datasource["requiresSecurityMaster"]))

    return available_datasets

@click.command(cls=LeanCommand, requires_lean_config=True, allow_unknown_options=True)
@click.option("--dataset", type=str, help="The name of the dataset to download non-interactively")
@click.option("--organization", type=str, help="The name or id of the organization to purchase and download data with")
@click.option("--overwrite", is_flag=True, default=False, help="Overwrite existing local data")
@click.pass_context
def download(ctx: click.Context,
             dataset: Optional[str],
             organization: Optional[str],
             overwrite: bool,
             **kwargs) -> None:
    """Purchase and download data from QuantConnect Datasets.

    An interactive wizard will show to walk you through the process of selecting data,
    accepting the CLI API Access and Data Agreement and payment.
    After this wizard the selected data will be downloaded automatically.

    If --dataset is given the command runs in non-interactive mode.
    In this mode the CLI does not prompt for input or confirmation but only halts when the agreement must be accepted.
    In non-interactive mode all options specific to the selected dataset as well as --organization are required.

    \b
    See the following url for the data that can be purchased and downloaded with this command:
    https://www.quantconnect.com/datasets
    """
    is_interactive = dataset is None and organization is None

    if not is_interactive:
        ensure_options(["dataset", "organization"])
        selected_organization = _get_organization_by_name_or_id(organization)
        datasets = _get_available_datasets(selected_organization)
        products = _select_products_non_interactive(selected_organization, datasets, ctx)
    else:
        selected_organization = _select_organization()
        datasets = _get_available_datasets(selected_organization)
        products = _select_products_interactive(selected_organization, datasets)

    _confirm_organization_balance(selected_organization, products)
    _verify_accept_agreement(selected_organization, is_interactive)

    if is_interactive:
        _confirm_payment(selected_organization, products)

    all_data_files = _get_data_files(selected_organization, products)
    container.data_downloader().download_files(all_data_files, overwrite, selected_organization.id)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from lean import __version__
from lean.components.util.click_aliased_command_group import AliasedCommandGroup


@click.group(cls=AliasedCommandGroup)
@click.version_option(__version__)
def lean() -> None:
    """The Lean CLI by QuantConnect."""
    # This method is intentionally empty
    # It is used as the command group for all `lean <command>` commands
    pass



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from datetime import datetime
from pathlib import Path
from typing import Optional
import click
from lean.click import LeanCommand, PathParameter
from lean.constants import DEFAULT_ENGINE_IMAGE, LEAN_ROOT_PATH
from lean.container import container, Logger
from lean.models.api import QCMinimalOrganization
from lean.models.utils import DebuggingMethod
from lean.models.logger import Option
from lean.models.data_providers import QuantConnectDataProvider, all_data_providers


# The _migrate_* methods automatically update launch configurations for a given debugging method.
#
# Occasionally we make changes which require updated launch configurations.
# Projects which are created after these update have the correct configuration already,
# but projects created before that need changes.
#
# These methods checks if the project has outdated configurations, and if so, update them to keep it working.

def _migrate_python_pycharm(logger: Logger, project_dir: Path) -> None:
    workspace_xml_path = project_dir / ".idea" / "workspace.xml"
    if not workspace_xml_path.is_file():
        return

    xml_manager = container.xml_manager()
    current_content = xml_manager.parse(workspace_xml_path.read_text(encoding="utf-8"))

    config = current_content.find('.//configuration[@name="Debug with Lean CLI"]')
    if config is None:
        return

    path_mappings = config.find('.//PathMappingSettings/option[@name="pathMappings"]/list')
    if path_mappings is None:
        return

    made_changes = False
    has_library_mapping = False

    library_dir = container.lean_config_manager().get_cli_root_directory() / "Library"
    if library_dir.is_dir():
        library_dir = f"$PROJECT_DIR$/{os.path.relpath(library_dir, project_dir)}".replace("\\", "/")
    else:
        library_dir = None

    for mapping in path_mappings.findall(".//mapping"):
        if mapping.get("local-root") == "$PROJECT_DIR$" and mapping.get("remote-root") == LEAN_ROOT_PATH:
            mapping.set("remote-root", "/LeanCLI")
            made_changes = True

        if library_dir is not None \
            and mapping.get("local-root") == library_dir \
            and mapping.get("remote-root") == "/Library":
            has_library_mapping = True

    if library_dir is not None and not has_library_mapping:
        library_mapping = xml_manager.parse("<mapping/>")
        library_mapping.set("local-root", library_dir)
        library_mapping.set("remote-root", "/Library")
        path_mappings.append(library_mapping)
        made_changes = True

    if made_changes:
        workspace_xml_path.write_text(xml_manager.to_string(current_content), encoding="utf-8")

        logger = container.logger()
        logger.warn("Your run configuration has been updated to work with the latest version of LEAN")
        logger.warn("Please restart the debugger in PyCharm and run this command again")

        raise click.Abort()


def _migrate_python_vscode(project_dir: Path) -> None:
    launch_json_path = project_dir / ".vscode" / "launch.json"
    if not launch_json_path.is_file():
        return

    current_content = json.loads(launch_json_path.read_text(encoding="utf-8"))
    if "configurations" not in current_content or not isinstance(current_content["configurations"], list):
        return

    config = next((c for c in current_content["configurations"] if c["name"] == "Debug with Lean CLI"), None)
    if config is None:
        return

    made_changes = False
    has_library_mapping = False

    library_dir = container.lean_config_manager().get_cli_root_directory() / "Library"
    if not library_dir.is_dir():
        library_dir = None

    for mapping in config["pathMappings"]:
        if mapping["localRoot"] == "${workspaceFolder}" and mapping["remoteRoot"] == LEAN_ROOT_PATH:
            mapping["remoteRoot"] = "/LeanCLI"
            made_changes = True

        if library_dir is not None and mapping["localRoot"] == str(library_dir) and mapping["remoteRoot"] == "/Library":
            has_library_mapping = True

    if library_dir is not None and not has_library_mapping:
        config["pathMappings"].append({
            "localRoot": str(library_dir),
            "remoteRoot": "/Library"
        })
        made_changes = True

    if made_changes:
        launch_json_path.write_text(json.dumps(current_content, indent=4), encoding="utf-8")


def _migrate_csharp_rider(logger: Logger, project_dir: Path) -> None:
    made_changes = False
    xml_manager = container.xml_manager()

    for dir_name in [f".idea.{project_dir.stem}", f".idea.{project_dir.stem}.dir"]:
        workspace_xml_path = project_dir / ".idea" / dir_name / ".idea" / "workspace.xml"
        if not workspace_xml_path.is_file():
            continue

        current_content = xml_manager.parse(workspace_xml_path.read_text(encoding="utf-8"))

        run_manager = current_content.find(".//component[@name='RunManager']")
        if run_manager is None:
            continue

        config = run_manager.find(".//configuration[@name='Debug with Lean CLI']")
        if config is None:
            continue

        run_manager.remove(config)

        workspace_xml_path.write_text(xml_manager.to_string(current_content), encoding="utf-8")
        made_changes = True

    if made_changes:
        container.project_manager().generate_rider_config()

        logger.warn("Your run configuration has been updated to work with the .NET 5 version of LEAN")
        logger.warn("Please restart Rider and start debugging again")
        logger.warn(
            "See https://www.lean.io/docs/v2/lean-cli/backtesting/debugging#05-C-and-Rider for the updated instructions")

        raise click.Abort()


def _migrate_csharp_vscode(project_dir: Path) -> None:
    launch_json_path = project_dir / ".vscode" / "launch.json"
    if not launch_json_path.is_file():
        return

    current_content = json.loads(launch_json_path.read_text(encoding="utf-8"))
    if "configurations" not in current_content or not isinstance(current_content["configurations"], list):
        return

    config = next((c for c in current_content["configurations"] if c["name"] == "Debug with Lean CLI"), None)
    if config is None:
        return

    if config["type"] != "mono" and config["processId"] != "${command:pickRemoteProcess}":
        return

    config.pop("address", None)
    config.pop("port", None)

    config["type"] = "coreclr"
    config["processId"] = "1"

    config["pipeTransport"] = {
        "pipeCwd": "${workspaceRoot}",
        "pipeProgram": "docker",
        "pipeArgs": ["exec", "-i", "lean_cli_vsdbg"],
        "debuggerPath": "/root/vsdbg/vsdbg",
        "quoteArgs": False
    }

    config["logging"] = {
        "moduleLoad": False
    }

    launch_json_path.write_text(json.dumps(current_content, indent=4), encoding="utf-8")


def _migrate_csharp_csproj(project_dir: Path) -> None:
    csproj_path = next((f for f in project_dir.rglob("*.csproj")), None)
    if csproj_path is None:
        return

    xml_manager = container.xml_manager()

    current_content = xml_manager.parse(csproj_path.read_text(encoding="utf-8"))
    if current_content.find(".//PropertyGroup/DefaultItemExcludes") is not None:
        return

    property_group = current_content.find(".//PropertyGroup")
    if property_group is None:
        property_group = xml_manager.parse("<PropertyGroup/>")
        current_content.append(property_group)

    default_item_excludes = xml_manager.parse(
        "<DefaultItemExcludes>$(DefaultItemExcludes);backtests/*/code/**;live/*/code/**;optimizations/*/code/**</DefaultItemExcludes>")
    property_group.append(default_item_excludes)

    csproj_path.write_text(xml_manager.to_string(current_content), encoding="utf-8")


def _select_organization() -> QCMinimalOrganization:
    """Asks the user for the organization that should be charged when downloading data.

    :return: the selected organization
    """
    api_client = container.api_client()

    organizations = api_client.organizations.get_all()
    options = [Option(id=organization, label=organization.name) for organization in organizations]

    logger = container.logger()
    return logger.prompt_list("Select the organization to purchase and download data with", options)


@click.command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@click.argument("project", type=PathParameter(exists=True, file_okay=True, dir_okay=True))
@click.option("--output",
              type=PathParameter(exists=False, file_okay=False, dir_okay=True),
              help="Directory to store results in (defaults to PROJECT/backtests/TIMESTAMP)")
@click.option("--detach", "-d",
              is_flag=True,
              default=False,
              help="Run the backtest in a detached Docker container and return immediately")
@click.option("--debug",
              type=click.Choice(["pycharm", "ptvsd", "vsdbg", "rider"], case_sensitive=False),
              help="Enable a certain debugging method (see --help for more information)")
@click.option("--data-provider",
              type=click.Choice([dp.get_name() for dp in all_data_providers], case_sensitive=False),
              help="Update the Lean configuration file to retrieve data from the given provider")
@click.option("--download-data",
              is_flag=True,
              default=False,
              help="Update the Lean configuration file to download data from the QuantConnect API, alias for --data-provider QuantConnect")
@click.option("--data-purchase-limit",
              type=int,
              help="The maximum amount of QCC to spend on downloading data during the backtest when using QuantConnect as data provider")
@click.option("--release",
              is_flag=True,
              default=False,
              help="Compile C# projects in release configuration instead of debug")
@click.option("--image",
              type=str,
              help=f"The LEAN engine image to use (defaults to {DEFAULT_ENGINE_IMAGE})")
@click.option("--python-venv",
              type=str,
              help=f"The path of the python virtual environment to be used")
@click.option("--update",
              is_flag=True,
              default=False,
              help="Pull the LEAN engine image before running the backtest")
@click.option("--backtest-name",
              type=str,
              help="Backtest name")
def backtest(project: Path,
             output: Optional[Path],
             detach: bool,
             debug: Optional[str],
             data_provider: Optional[str],
             download_data: bool,
             data_purchase_limit: Optional[int],
             release: bool,
             image: Optional[str],
             python_venv: Optional[str],
             update: bool,
             backtest_name: str) -> None:
    """Backtest a project locally using Docker.

    \b
    If PROJECT is a directory, the algorithm in the main.py or Main.cs file inside it will be executed.
    If PROJECT is a file, the algorithm in the specified file will be executed.

    \b
    Go to the following url to learn how to debug backtests locally using the Lean CLI:
    https://www.lean.io/docs/v2/lean-cli/backtesting/debugging

    By default the official LEAN engine image is used.
    You can override this using the --image option.
    Alternatively you can set the default engine image for all commands using `lean config set engine-image <image>`.
    """
    logger = container.logger()
    project_manager = container.project_manager()
    algorithm_file = project_manager.find_algorithm_file(Path(project))
    lean_config_manager = container.lean_config_manager()

    if output is None:
        output = algorithm_file.parent / "backtests" / datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    debugging_method = None
    if debug == "pycharm":
        debugging_method = DebuggingMethod.PyCharm
        _migrate_python_pycharm(logger, algorithm_file.parent)
    elif debug == "ptvsd":
        debugging_method = DebuggingMethod.PTVSD
        _migrate_python_vscode(algorithm_file.parent)
    elif debug == "vsdbg":
        debugging_method = DebuggingMethod.VSDBG
        _migrate_csharp_vscode(algorithm_file.parent)
    elif debug == "rider":
        debugging_method = DebuggingMethod.Rider
        _migrate_csharp_rider(logger, algorithm_file.parent)

    if debugging_method is not None and detach:
        raise RuntimeError("Running a debugging session in a detached container is not supported")

    if algorithm_file.name.endswith(".cs"):
        _migrate_csharp_csproj(algorithm_file.parent)

    lean_config = lean_config_manager.get_complete_lean_config("backtesting", algorithm_file, debugging_method)

    if download_data:
        data_provider = QuantConnectDataProvider.get_name()

    if data_provider is not None:
        data_provider = next(dp for dp in all_data_providers if dp.get_name() == data_provider)
        data_provider.build(lean_config, logger).configure(lean_config, "backtesting")

    lean_config_manager.configure_data_purchase_limit(lean_config, data_purchase_limit)

    cli_config_manager = container.cli_config_manager()
    project_config_manager = container.project_config_manager()

    project_config = project_config_manager.get_project_config(algorithm_file.parent)
    engine_image = cli_config_manager.get_engine_image(image or project_config.get("engine-image", None))

    if str(engine_image) != DEFAULT_ENGINE_IMAGE:
        logger.warn(f'A custom engine image: "{engine_image}" is being used!')

    container.update_manager().pull_docker_image_if_necessary(engine_image, update)

    if not output.exists():
        output.mkdir(parents=True)

    output_config_manager = container.output_config_manager()
    lean_config["algorithm-id"] = str(output_config_manager.get_backtest_id(output))

    # Set backtest name
    if backtest_name is not None and backtest_name != "":
        lean_config["backtest-name"] = backtest_name

    if python_venv is not None and python_venv != "":
        lean_config["python-venv"] = f'{"/" if python_venv[0] != "/" else ""}{python_venv}'

    lean_runner = container.lean_runner()
    lean_runner.run_lean(lean_config,
                         "backtesting",
                         algorithm_file,
                         output,
                         engine_image,
                         debugging_method,
                         release,
                         detach)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dependency_injector.containers import DeclarativeContainer
from dependency_injector.providers import Factory, Singleton

from lean.components.api.api_client import APIClient
from lean.components.cloud.cloud_project_manager import CloudProjectManager
from lean.components.cloud.cloud_runner import CloudRunner
from lean.components.cloud.data_downloader import DataDownloader
from lean.components.cloud.module_manager import ModuleManager
from lean.components.cloud.pull_manager import PullManager
from lean.components.cloud.push_manager import PushManager
from lean.components.config.cli_config_manager import CLIConfigManager
from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.config.optimizer_config_manager import OptimizerConfigManager
from lean.components.config.output_config_manager import OutputConfigManager
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.config.storage import Storage
from lean.components.docker.docker_manager import DockerManager
from lean.components.docker.lean_runner import LeanRunner
from lean.components.util.http_client import HTTPClient
from lean.components.util.library_manager import LibraryManager
from lean.components.util.logger import Logger
from lean.components.util.market_hours_database import MarketHoursDatabase
from lean.components.util.name_generator import NameGenerator
from lean.components.util.path_manager import PathManager
from lean.components.util.platform_manager import PlatformManager
from lean.components.util.project_manager import ProjectManager
from lean.components.util.task_manager import TaskManager
from lean.components.util.temp_manager import TempManager
from lean.components.util.update_manager import UpdateManager
from lean.components.util.xml_manager import XMLManager
from lean.constants import CACHE_PATH, CREDENTIALS_CONFIG_PATH, GENERAL_CONFIG_PATH


class Container(DeclarativeContainer):
    """The Container class wires all reusable components together."""
    logger = Singleton(Logger)

    platform_manager = Singleton(PlatformManager)
    task_manager = Singleton(TaskManager, logger)
    name_generator = Singleton(NameGenerator)
    path_manager = Singleton(PathManager, platform_manager)
    temp_manager = Singleton(TempManager)
    xml_manager = Singleton(XMLManager)
    http_client = Singleton(HTTPClient, logger)

    general_storage = Singleton(Storage, file=GENERAL_CONFIG_PATH)
    credentials_storage = Singleton(Storage, file=CREDENTIALS_CONFIG_PATH)
    cache_storage = Singleton(Storage, file=CACHE_PATH)

    cli_config_manager = Singleton(CLIConfigManager, general_storage, credentials_storage)

    api_client = Factory(APIClient,
                         logger,
                         http_client,
                         user_id=cli_config_manager.provided.user_id.get_value()(),
                         api_token=cli_config_manager.provided.api_token.get_value()())

    module_manager = Singleton(ModuleManager, logger, api_client, http_client)

    project_config_manager = Singleton(ProjectConfigManager, xml_manager)
    lean_config_manager = Singleton(LeanConfigManager,
                                    logger,
                                    cli_config_manager,
                                    project_config_manager,
                                    module_manager,
                                    cache_storage)
    output_config_manager = Singleton(OutputConfigManager, lean_config_manager)
    optimizer_config_manager = Singleton(OptimizerConfigManager, logger)

    project_manager = Singleton(ProjectManager,
                                logger,
                                project_config_manager,
                                lean_config_manager,
                                path_manager,
                                xml_manager,
                                platform_manager)
    library_manager = Singleton(LibraryManager,
                                logger,
                                project_manager,
                                project_config_manager,
                                lean_config_manager,
                                path_manager,
                                xml_manager)

    cloud_runner = Singleton(CloudRunner, logger, api_client, task_manager)
    pull_manager = Singleton(PullManager,
                             logger,
                             api_client,
                             project_manager,
                             project_config_manager,
                             library_manager,
                             platform_manager)
    push_manager = Singleton(PushManager, logger, api_client, project_manager, project_config_manager)
    data_downloader = Singleton(DataDownloader, logger, api_client, lean_config_manager)
    cloud_project_manager = Singleton(CloudProjectManager,
                                      api_client,
                                      project_config_manager,
                                      pull_manager,
                                      push_manager,
                                      path_manager)

    docker_manager = Singleton(DockerManager, logger, temp_manager, platform_manager)
    lean_runner = Singleton(LeanRunner,
                            logger,
                            project_config_manager,
                            lean_config_manager,
                            output_config_manager,
                            docker_manager,
                            module_manager,
                            project_manager,
                            temp_manager,
                            xml_manager)

    market_hours_database = Singleton(MarketHoursDatabase, lean_config_manager)

    update_manager = Singleton(UpdateManager, logger, http_client, cache_storage, docker_manager)


container = Container()
container.data_downloader().update_database_files()



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from lean.models.pydantic import WrappedBaseModel


class Option(WrappedBaseModel):
    """The Option class represents a choosable option with an internal id and a display-friendly label."""
    id: Any
    label: str



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import requests

from lean.components.util.logger import Logger
from lean.models.errors import MoreInfoError


class HTTPClient:
    """The HTTPClient class is a lightweight wrapper around the requests library with additional logging."""

    def __init__(self, logger: Logger) -> None:
        """Creates a new HTTPClient instance.

        :param logger: the logger to log debug messages with
        """
        self._logger = logger

    def get(self, url: str, **kwargs) -> requests.Response:
        """A wrapper around requests.get().

        An error is raised if the response is unsuccessful unless kwargs["raise_for_status"] == False.

        :param url: the request url
        :param kwargs: any kwargs to pass on to requests.get()
        :return: the response of the request
        """
        return self.request("GET", url, **kwargs)

    def post(self, url: str, **kwargs) -> requests.Response:
        """A wrapper around requests.post().

        An error is raised if the response is unsuccessful unless kwargs["raise_for_status"] == False.

        :param url: the request url
        :param kwargs: any kwargs to pass on to requests.post()
        :return: the response of the request
        """
        return self.request("POST", url, **kwargs)

    def request(self, method: str, url: str, **kwargs) -> requests.Response:
        """A wrapper around requests.request().

        An error is raised if the response is unsuccessful unless kwargs["raise_for_status"] == False.

        :param method: the request method
        :param url: the request url
        :param kwargs: any kwargs to pass on to requests.request()
        :return: the response of the request
        """
        self._log_request(method, url, **kwargs)

        raise_for_status = kwargs.pop("raise_for_status", True)
        try:
            response = requests.request(method, url, **kwargs)
        except requests.exceptions.SSLError as e:
            raise Exception(f"""
Detected SSL error, this might be due to custom certificates in your environment or system trust store.
A known limitation of the python requests implementation. 
Please consider installing library https://pypi.org/project/python-certifi-win32/. 
Related issue https://github.com/psf/requests/issues/2966
    """.strip())

        self._check_response(response, raise_for_status)
        return response

    def log_unsuccessful_response(self, response: requests.Response) -> None:
        """Logs an unsuccessful response's status code and body.

        :param response: the response to log
        """
        body = f"body:\n{response.text}" if response.text != "" else "empty body"
        self._logger.debug(f"Request was not successful, status code {response.status_code}, {body}")

    def _log_request(self, method: str, url: str, **kwargs) -> None:
        """Logs a request.

        :param method: the request method
        :param url: the request url
        :param kwargs: any kwargs passed to a request.* method
        """
        message = f"--> {method.upper()} {url}"

        data = next((kwargs.get(key) for key in ["json", "data", "params"] if key in kwargs), None)
        if data is not None and data != {}:
            message += f" with data:\n{json.dumps(data, indent=4)}"

        self._logger.debug(message)

    def _check_response(self, response: requests.Response, raise_for_status: bool) -> None:
        """Checks a response, logging a debug message if it wasn't successful.

        :param response: the response to check
        :param raise_for_status: True if an error needs to be raised if the request wasn't successful, False if not
        """
        if response.status_code < 200 or response.status_code >= 300:
            self.log_unsuccessful_response(response)

        if raise_for_status:
            response.raise_for_status()
